from typing import List, Optional

import pytest  # type:ignore[import]

from maus import group_lines_by_segment_group, merge_lines_with_same_data_element, to_deep_ahb
from maus.models.anwendungshandbuch import AhbLine, AhbMetaInformation, DeepAnwendungshandbuch, FlatAnwendungshandbuch
from maus.models.edifact_components import (
    DataElement,
    DataElementFreeText,
    DataElementValuePool,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
)
from maus.models.message_implementation_guide import SegmentGroupHierarchy
from unittests.serialization_test_helper import assert_serialization_roundtrip  # type:ignore[import]


class TestMaus:
    """
    Tests the behaviour of the MAUS.
    """

    @pytest.mark.parametrize(
        "ahb_lines, expected_data_element",
        [
            pytest.param(
                # The following two lines should be merged into a single data element value pool because they
                # refer to the same data element but have different value pool entries.
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                DataElementValuePool(
                    discriminator="SG4->FOO->0333",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="E01",
                            meaning="Das Eine",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="E02",
                            meaning="Das Andere",
                            ahb_expression="X",
                        ),
                    ],
                    data_element_id="0333",
                ),
            ),
            pytest.param(
                # The following single lines should result in a single DataElement with free text option
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry=None,
                        name="Das Eine",
                        guid=None,
                    ),
                ],
                DataElementFreeText(
                    discriminator="Das Eine",
                    ahb_expression="X",
                    entered_input=None,
                    data_element_id="0333",
                ),
            ),
        ],
    )
    def test_data_element_merging(self, ahb_lines: List[AhbLine], expected_data_element: DataElement):
        actual = merge_lines_with_same_data_element(ahb_lines)
        assert actual == expected_data_element

    @pytest.mark.parametrize(
        "ahb_lines",
        [
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0334",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                id="lines with different data element",
            ),
            pytest.param([], id="empty list"),
        ],
    )
    def test_data_element_value_error(self, ahb_lines: List[AhbLine]):
        with pytest.raises(ValueError):
            merge_lines_with_same_data_element(ahb_lines)

    @pytest.mark.parametrize(
        "ahb_lines, expected_data_element",
        [
            pytest.param(
                # The following two lines should be merged into a single data element value pool because they
                # refer to the same data element but have different value pool entries.
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                DataElementValuePool(
                    discriminator="SG4->FOO->0333",
                    value_pool=[
                        ValuePoolEntry(
                            qualifier="E01",
                            meaning="Das Eine",
                            ahb_expression="X",
                        ),
                        ValuePoolEntry(
                            qualifier="E02",
                            meaning="Das Andere",
                            ahb_expression="X",
                        ),
                    ],
                    data_element_id="0333",
                ),
            ),
            pytest.param(
                # The following single lines should result in a single DataElement with free text option
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry=None,
                        name="Das Eine",
                        guid=None,
                    ),
                ],
                DataElementFreeText(
                    discriminator="Das Eine",
                    ahb_expression="X",
                    entered_input=None,
                    data_element_id="0333",
                ),
            ),
        ],
    )
    def test_nest_segment_groups_into_each_other(self, ahb_lines: List[AhbLine], expected_data_element: DataElement):
        actual = merge_lines_with_same_data_element(ahb_lines)
        assert actual == expected_data_element

    @pytest.mark.parametrize(
        "ahb_lines, sgh, expected_result",
        [
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                SegmentGroupHierarchy(segment_group="SG4", sub_hierarchy=None, opening_segment="FOO"),
                [
                    SegmentGroup(
                        discriminator="SG4",
                        ahb_expression="X",
                        segments=[
                            Segment(
                                discriminator="FOO",
                                ahb_expression="X",
                                data_elements=[
                                    DataElementValuePool(
                                        discriminator="SG4->FOO->0333",
                                        value_pool=[
                                            ValuePoolEntry(
                                                qualifier="E01",
                                                meaning="Das Eine",
                                                ahb_expression="X",
                                            ),
                                            ValuePoolEntry(
                                                qualifier="E02",
                                                meaning="Das Andere",
                                                ahb_expression="X",
                                            ),
                                        ],
                                        data_element_id="0333",
                                    )
                                ],
                            )
                        ],
                        segment_groups=[],
                    )
                ],
            ),
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E01",
                        name="Das Eine",
                        guid=None,
                    ),
                    AhbLine(
                        segment_group_key="SG4",
                        segment_code="FOO",
                        ahb_expression="X",
                        data_element="0333",
                        value_pool_entry="E02",
                        name="Das Andere",
                        guid=None,
                    ),
                ],
                SegmentGroupHierarchy(segment_group="SG4", sub_hierarchy=None, opening_segment="FOO"),
                [
                    SegmentGroup(
                        discriminator="SG4",
                        ahb_expression="X",
                        segments=[
                            Segment(
                                discriminator="FOO",
                                ahb_expression="X",
                                data_elements=[
                                    DataElementValuePool(
                                        discriminator="SG4->FOO->0333",
                                        value_pool=[
                                            ValuePoolEntry(
                                                qualifier="E01",
                                                meaning="Das Eine",
                                                ahb_expression="X",
                                            ),
                                            ValuePoolEntry(
                                                qualifier="E02",
                                                meaning="Das Andere",
                                                ahb_expression="X",
                                            ),
                                        ],
                                        data_element_id="0333",
                                    )
                                ],
                            )
                        ],
                        segment_groups=[],
                    )
                ],
            ),
        ],
    )
    def test_group_lines_by_segment_group(
        self, ahb_lines: List[AhbLine], sgh: SegmentGroupHierarchy, expected_result: List[SegmentGroup]
    ):
        actual = group_lines_by_segment_group(ahb_lines, sgh)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "sgh, flat_ahb, expected_deep",
        [
            pytest.param(
                SegmentGroupHierarchy(
                    segment_group=None,
                    opening_segment="UNH",
                    sub_hierarchy=[
                        SegmentGroupHierarchy(
                            segment_group="SG4",
                            opening_segment="FOO",
                            sub_hierarchy=[
                                SegmentGroupHierarchy(
                                    segment_group="SG5",
                                    opening_segment="BAR",
                                    sub_hierarchy=[
                                        SegmentGroupHierarchy(
                                            segment_group="SG6", opening_segment="BAZ", sub_hierarchy=None
                                        )
                                    ],
                                ),
                                SegmentGroupHierarchy(
                                    segment_group="SG7",
                                    opening_segment="ASD",
                                    sub_hierarchy=[
                                        SegmentGroupHierarchy(
                                            segment_group="SG8", opening_segment="YXC", sub_hierarchy=None
                                        )
                                    ],
                                ),
                            ],
                        )
                    ],
                ),
                FlatAnwendungshandbuch(
                    meta=AhbMetaInformation(pruefidentifikator="12345"),
                    lines=[
                        # this segment should stay at the beginning, alone
                        AhbLine(
                            segment_group_key=None,
                            segment_code="UNH",
                            ahb_expression="X",
                            data_element="1234",
                            value_pool_entry=None,
                            name="Nachrichten-Startsegment",
                            guid=None,
                        ),
                        # The following two lines should be merged into a single data element value pool because they
                        # refer to the same data element but have different value pool entries.
                        AhbLine(
                            segment_group_key="SG4",
                            segment_code="FOO",
                            ahb_expression="X",
                            data_element="0333",
                            value_pool_entry="E01",
                            name="Das andere",
                            guid=None,
                        ),
                        AhbLine(
                            segment_group_key="SG4",
                            segment_code="FOO",
                            ahb_expression="X",
                            data_element="0333",
                            value_pool_entry="E02",
                            name="Das Eine",
                            guid=None,
                        ),
                        AhbLine(
                            segment_group_key="SG5",
                            segment_code="BAR",
                            ahb_expression="X",
                            data_element="1234",
                            name="Die fünfte Gruppe",
                            guid=None,
                            value_pool_entry=None,
                        ),
                        AhbLine(
                            segment_group_key="SG6",
                            segment_code="irgendwas",
                            ahb_expression="X",
                            data_element="7889",
                            name="Die sechste Gruppe",
                            guid=None,
                            value_pool_entry=None,
                        ),
                        AhbLine(
                            segment_group_key="SG7",
                            segment_code="ASD",
                            ahb_expression="X",
                            data_element="0123",
                            name="Die siebte Gruppe",
                            guid=None,
                            value_pool_entry=None,
                        ),
                        # die achte fehlt
                    ],
                ),
                DeepAnwendungshandbuch(
                    meta=AhbMetaInformation(pruefidentifikator="12345"),
                    lines=[
                        SegmentGroup(
                            discriminator="root",
                            ahb_expression="X",
                            segments=[
                                Segment(
                                    discriminator="UNH",
                                    ahb_expression="X",
                                    data_elements=[
                                        DataElementFreeText(
                                            discriminator="Nachrichten-Startsegment",
                                            ahb_expression="X",
                                            entered_input=None,
                                            data_element_id="1234",
                                        )
                                    ],
                                )
                            ],
                            segment_groups=[
                                SegmentGroup(
                                    discriminator="SG4",
                                    ahb_expression="X",
                                    segments=[
                                        Segment(
                                            discriminator="FOO",
                                            ahb_expression="X",
                                            data_elements=[
                                                DataElementValuePool(
                                                    discriminator="SG4->FOO->0333",
                                                    value_pool=[
                                                        ValuePoolEntry(
                                                            qualifier="E01",
                                                            meaning="Das andere",
                                                            ahb_expression="X",
                                                        ),
                                                        ValuePoolEntry(
                                                            qualifier="E02",
                                                            meaning="Das Eine",
                                                            ahb_expression="X",
                                                        ),
                                                    ],
                                                    data_element_id="0333",
                                                )
                                            ],
                                        )
                                    ],
                                    segment_groups=[
                                        SegmentGroup(
                                            discriminator="SG5",
                                            ahb_expression="X",
                                            segments=[
                                                Segment(
                                                    discriminator="BAR",
                                                    ahb_expression="X",
                                                    data_elements=[
                                                        DataElementFreeText(
                                                            discriminator="Die fünfte Gruppe",
                                                            ahb_expression="X",
                                                            entered_input=None,
                                                            data_element_id="1234",
                                                        )
                                                    ],
                                                )
                                            ],
                                            segment_groups=[
                                                SegmentGroup(
                                                    discriminator="SG6",
                                                    ahb_expression="X",
                                                    segments=[
                                                        Segment(
                                                            discriminator="irgendwas",
                                                            ahb_expression="X",
                                                            data_elements=[
                                                                DataElementFreeText(
                                                                    discriminator="Die sechste Gruppe",
                                                                    ahb_expression="X",
                                                                    entered_input=None,
                                                                    data_element_id="7889",
                                                                )
                                                            ],
                                                        )
                                                    ],
                                                    segment_groups=[],
                                                )
                                            ],
                                        ),
                                        SegmentGroup(
                                            discriminator="SG7",
                                            ahb_expression="X",
                                            segments=[
                                                Segment(
                                                    discriminator="ASD",
                                                    ahb_expression="X",
                                                    data_elements=[
                                                        DataElementFreeText(
                                                            discriminator="Die siebte Gruppe",
                                                            ahb_expression="X",
                                                            entered_input=None,
                                                            data_element_id="0123",
                                                        )
                                                    ],
                                                )
                                            ],
                                            segment_groups=[],
                                        ),
                                    ],
                                )
                            ],
                        )
                    ],
                ),
            ),
        ],
    )
    def test_ahb_flat_to_deep(
        self, sgh: SegmentGroupHierarchy, flat_ahb: FlatAnwendungshandbuch, expected_deep: DeepAnwendungshandbuch
    ):
        actual = to_deep_ahb(flat_ahb, sgh)
        # actual_json = DeepAnwendungshandbuchSchema().dumps(actual)
        assert actual == expected_deep



from pathlib import Path
from typing import Optional

import pytest  # type:ignore[import]
from lxml import etree  # type:ignore[import]

from maus.models.edifact_components import EdifactStackQuery
from maus.reader.mig_reader import MigXmlReader

ALL_MIG_XML_FILES = pytest.mark.datafiles(
    "./unittests/migs/FV2204/template_xmls/mscons_1154.xml",
    "./unittests/migs/FV2204/template_xmls/utilmd_1154.xml",
    "./unittests/migs/FV2204/template_xmls/utilmd_2379.xml",
    "./unittests/migs/FV2204/template_xmls/utilmd_2380.xml",
    "./unittests/migs/FV2204/template_xmls/utilmd_7402.xml",
    "./unittests/migs/FV2204/template_xmls/utilmd_3225.xml",
    "./unittests/migs/FV2204/template_xmls/utilmd_9013.xml",
    "./unittests/migs/FV2204/template_xmls/reqote.xml",
    # "./unittests/migs/FV2204/template_xmls/utilmd.xml",
    # "./unittests/migs/FV2204/template_xmls/iftsta.xml",
)


class TestMigXmlReader:
    """
    Tests the behaviour of the Message Implementation Guide model
    """

    @pytest.mark.parametrize(
        "xml_string, expected_format",
        [
            pytest.param('<?xml version="1.0"?><MSCONS></MSCONS>', "MSCONS"),
            pytest.param('<?xml version="1.0"?><UTILMD></UTILMD>', "UTILMD"),
        ],
    )
    def test_get_edifact_format(self, xml_string: str, expected_format: str):
        reader = MigXmlReader(xml_string)
        assert reader.get_format_name() == expected_format

    @pytest.mark.parametrize(
        "x,y, expected_result",
        [
            pytest.param(None, "", True),
            pytest.param("", "", True),
            pytest.param("X", "X", True),
            pytest.param("X", "x", True),
            pytest.param("X", "y", False),
            pytest.param("Gültigkeit, Beginndatum", "Gültigkeit,Beginndatum", True),
            pytest.param(
                "Referenz Vorgangsnummer (aus Anfragenachricht)", "Referenz Vorgangsnummer (aus Anfragenachricht)", True
            ),
        ],
    )
    def test_are_similar_names(self, x: Optional[str], y: Optional[str], expected_result: bool):
        actual = MigXmlReader.are_similar_names(x, y)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "segment_code,expected_is_boilerplate",
        [
            pytest.param(None, True),
            pytest.param("UNH", False),
            pytest.param("UNT", True),
            pytest.param("UNZ", True),
            pytest.param("DTM", False),
        ],
    )
    def test_is_boilerplate_segment(self, segment_code: Optional[str], expected_is_boilerplate: bool):
        actual_is_boilerplate = MigXmlReader.is_edifact_boilerplate(segment_code)
        assert actual_is_boilerplate == expected_is_boilerplate

    def test_make_tree_names_comparable(self):
        orig_xml = """<?xml version="1.0"?>
        <hello>
        <einfield name="Foo "/>
        <einclass ahbName="Hallo-Welt">
            <einfield name=" DiGiTaLiSiErUnG" ahbName="(ZuKunFt)"/>
        </einclass>
        </hello>
        """
        tree = etree.ElementTree(etree.fromstring(orig_xml))
        MigXmlReader.make_tree_names_comparable(tree)
        for element in tree.iter():
            if "name" in element.attrib:
                assert element.attrib["name"] == MigXmlReader.make_name_comparable(element.attrib["name"])
            if "ahbName" in element.attrib:
                assert element.attrib["ahbName"] == MigXmlReader.make_name_comparable(element.attrib["ahbName"])

    @ALL_MIG_XML_FILES
    @pytest.mark.parametrize(
        "mig_xml_path, query, expected_path",
        [
            pytest.param(
                "mscons_1154.xml",
                EdifactStackQuery(
                    segment_group_key="SG1", segment_code="RFF", data_element_id="1154", name="Prüfidentifikator"
                ),
                '$["Dokument"][0]["Nachricht"][0]["Prüfidentifikator"][0]["ID"]',
                id="mscons pruefi",
            ),
            pytest.param(
                "utilmd_7402.xml",
                EdifactStackQuery(
                    segment_group_key="SG4", segment_code="IDE", data_element_id="7402", name="Vorgangsnummer"
                ),
                '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Vorgangsnummer"]',
                id="utilmd vorgangsnummer",
            ),
            pytest.param(
                "utilmd_2380.xml",
                EdifactStackQuery(
                    segment_group_key="root", segment_code="DTM", data_element_id="2380", name="Gültigkeit, Beginndatum"
                ),
                '$["Dokument"][0]["Nachricht"][0]["Gültigkeit,Beginndatum"]',
                id="gueltigkeit,beginndatum",
            ),
            pytest.param(
                "utilmd_2380.xml",
                EdifactStackQuery(
                    segment_group_key="SG4", segment_code="DTM", data_element_id="2380", name="Bilanzierungsbeginn"
                ),
                '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Bilanzierungsbeginn"]',
                id="utilmd bilanzierungsbeginn",
            ),
            pytest.param(
                "utilmd_1154.xml",
                EdifactStackQuery(
                    segment_group_key="SG6",
                    segment_code="RFF",
                    data_element_id="1154",
                    name="Referenz Vorgangsnummer(aus Anfragenachricht)",
                ),
                '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Referenz Vorgangsnummer (aus Anfragenachricht)"][0]["Referenz"]',
                id="referenz aus anfragenachricht",
            ),
            pytest.param(
                "utilmd_1154.xml",
                EdifactStackQuery(
                    segment_group_key="SG8",
                    segment_code="RFF",
                    data_element_id="1154",
                    name="Referenz auf die ID einer Messlokation",
                    predecessor_qualifier="Z19",
                ),
                '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Erforderliche OBIS-Daten der Messlokation"][0]["Referenz auf die ID einer Messlokation"]',
                id="referenz auf melo-id",
            ),
            pytest.param(
                "reqote.xml",
                EdifactStackQuery(
                    segment_group_key="root",
                    segment_code="DTM",
                    data_element_id="2380",
                    name="Datum oder Uhrzeit oderZeitspanne, Wert",
                    # <-- the messed up spaces are due to line breaks in the PDF
                    predecessor_qualifier="76",
                ),
                '$["Dokument"][0]["Nachricht"][0]["Datum zum geplanten Leistungsbeginn"]',
                id="reqote lieferdatum",
            ),
            pytest.param(
                "reqote.xml",
                EdifactStackQuery(
                    segment_group_key="SG11",
                    segment_code="NAD",
                    data_element_id="3039",
                    name="MP-ID",  # <-- this used to be "Beteiligter, Identifikation"... don't know why
                    predecessor_qualifier="MS",
                ),
                '$["Dokument"][0]["Nachricht"][0]["MP-ID Absender"][0]["MP-ID"]',
                id="reqote absender",
            ),
            pytest.param(
                "reqote.xml",
                EdifactStackQuery(
                    segment_group_key="SG11",
                    segment_code="NAD",
                    data_element_id="3055",
                    name=None,
                    predecessor_qualifier="MS",
                ),
                '$["Dokument"][0]["Nachricht"][0]["MP-ID Absender"][0]["Codeliste"]',
                id="reqote absender codeliste",
            ),
            pytest.param(
                "utilmd_9013.xml",
                EdifactStackQuery(
                    segment_group_key="SG4",
                    segment_code="STS",
                    data_element_id="9013",
                    name=None,
                    predecessor_qualifier="7",
                ),
                '$["Dokument"][0]["Vorgang"][0]["Transaktionsgrund"]',
                id="UTILMD Transaktionsgrund",
            ),
            pytest.param(
                "utilmd_3225.xml",
                EdifactStackQuery(
                    segment_group_key="SG5",
                    segment_code="LOC",
                    data_element_id="3225",
                    name="Identifikator",  # <-- name in AHB != name in MIG ("ID")
                    predecessor_qualifier="172",
                ),
                '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Meldepunkt"][0]["ID"]',
                id="UTILMD: LOC ID vs. Identifikator",
            ),
            pytest.param(
                "utilmd_1154.xml",
                EdifactStackQuery(
                    segment_group_key="SG6",
                    segment_code="RFF",  # this is a reference
                    data_element_id="1154",
                    name="ID der Marktlokation",  # <-- name in AHB != name in MIG ("ID")
                    predecessor_qualifier="Z18",
                ),
                '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Referenz auf die ID der Marktlokation für Termine der Marktlokation"][0]["ID"]',
                id="UTILMD: LOC ID der MaLo Z18",
            ),
            pytest.param(
                "utilmd_2380.xml",
                EdifactStackQuery(
                    segment_group_key="SG6",
                    segment_code="DTM",
                    data_element_id="2380",
                    name="Datum oder Uhrzeit oderZeitspanne, Wert",
                    predecessor_qualifier="752",
                    section_name="Geplante Turnusablesung des MSB (Strom)"
                    # division=Division.ELECTRICITY,
                ),
                '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Geplante Turnusablesung des MSB (Strom)"]',
                id="UTILMD geplante Turnusablesung 752",
            ),
            # pytest.param( # unsolved
            #    "utilmd_1154.xml",
            # EdifactStackQuery(segment_group_key='SG8', segment_code='RFF', data_element_id='1154',
            #                  name='ID der Marktlokation', predecessor_qualifier='Z18',
            #                  section_name='Referenz auf die ID der Marktlokation'),
            #    '$["Dokument"][0]["Vorgang"]["0"]["Daten der Marktlokation"][0]["Referenz auf die ID der Marktlokation"]',
            #    id="Referenz auf die ID der Marktlokation"
            # )
            # pytest.param( # unsolved
            #    "utilmd_2379.xml",
            #    EdifactStackQuery(
            #        segment_group_key="SG6",
            #        segment_code="DTM",
            #        data_element_id="2379",
            #        name=None,
            #        predecessor_qualifier="752",
            #        section_name="Geplante Turnusablesung des MSB (Strom)",
            #    ),
            #    "$['foo']",
            #    id="UTILMD geplante Turnusablesung (Qualifier)",
            #    # https://github.com/Hochfrequenz/edifact-templates/issues/24
            # ),
        ],
    )
    def test_simple_paths(
        self,
        datafiles,
        mig_xml_path: str,
        query: EdifactStackQuery,
        expected_path: str,
    ):
        reader = MigXmlReader(Path(datafiles) / mig_xml_path)
        actual_stack = reader.get_edifact_stack(query)
        assert actual_stack is not None
        assert actual_stack.to_json_path() == expected_path  # type:ignore[union-attr]



# pylint:disable=too-few-public-methods
"""
EDIFACT components are data structures on different hierarchical levels inside an EDIFACT message.
Components contain not only EDIFACT composits but also segments and segment groups.
"""
import re
from abc import ABC
from enum import Enum
from typing import List, Optional, Type

import attr
import attrs
from marshmallow import Schema, fields, post_dump, post_load, pre_dump, pre_load  # type:ignore[import]
from marshmallow_enum import EnumField  # type:ignore[import]


class DataElementDataType(str, Enum):
    """
    The Data Element Data Type describes with which kind of data element we're dealing with in a data element.
    This information is set but not used anywhere inside MAUS directly but more of a "service" to downstream code.
    """

    TEXT = "TEXT"  #: plain text, f.e. a name
    DATETIME = "DATETIME"  #: a datetime string, usually as RFC3339
    VALUE_POOL = "VALUE_POOL"  #: the user can choose between different possible values


def derive_data_type_from_segment_code(segment_code: str) -> Optional[DataElementDataType]:
    """
    derives the expected data type from the segment code, f.e. `DATETIME` for DTM segments
    :return: The DataType if it can be derived without any doubt, None otherwise
    """
    if segment_code in {"DTM"}:
        return DataElementDataType.DATETIME
    return None


@attrs.define(auto_attribs=True, kw_only=True)
class DataElement(ABC):
    """
    A data element holds specific kinds of data. It is defined in EDIFACT.
    At least for the German energy market communication any data element has a 4 digit key.
    For example in UTILMD the data element that holds the 13 digit market partner ID is data element '3039'
    """

    discriminator: str = attrs.field(validator=attrs.validators.instance_of(str))
    """ The discriminator uniquely identifies the data element. This _might_ be its key """
    # but could also be a reference or a name
    #: the ID of the data element (f.e. "0062") for the Nachrichten-Referenznummer
    data_element_id: str = attrs.field(validator=attrs.validators.matches_re(r"^\d{4}$"))
    #: the type of data expected to be used with this data element
    value_type: Optional[DataElementDataType] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(DataElementDataType)), default=None
    )


class DataElementSchema(Schema):
    """
    A Schema to (de-)serialize DataElements
    """

    discriminator = fields.String(required=True)
    data_element_id = fields.String(required=True)
    value_type = EnumField(DataElementDataType, required=False)


@attrs.define(auto_attribs=True, kw_only=True)
class DataElementFreeText(DataElement):
    """
    A DataElementFreeText is a data element that allows entering arbitrary data.
    This is the main difference to the :class:`DataElementValuePool` which has a finite set of allowed values attached.
    """

    value_type: Optional[DataElementDataType] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(DataElementDataType)),  # type:ignore[arg-type]
        default=DataElementDataType.TEXT,
    )
    ahb_expression: str = attrs.field(validator=attrs.validators.instance_of(str))
    """any freetext data element has an ahb expression attached. Could be 'X' but also 'M [13]'"""
    entered_input: Optional[str] = attrs.field(validator=attrs.validators.optional(attrs.validators.instance_of(str)))
    """If the message contains data for this data element, this is not None."""


class DataElementFreeTextSchema(DataElementSchema):
    """
    A Schema to serialize DataElementFreeTexts
    """

    ahb_expression = fields.String(required=True)
    entered_input = fields.String(required=False, load_default=None)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> DataElementFreeText:
        """
        Converts the barely typed data dictionary into an actual :class:`.DataElementFreeText`
        """
        return DataElementFreeText(**data)


# pylint: disable=unused-argument
def _check_that_string_is_not_whitespace_or_empty(instance, attribute, value):
    """
    Check that string in the instance attribute value is not empty
    """
    if not value:
        raise ValueError(f"The string {attribute.name} must not be None or empty")
    if len(value.strip()) == 0:
        raise ValueError(f"The string {attribute.name} must not consist only of whitespace: '{value}'")


@attrs.define(auto_attribs=True, kw_only=True)
class ValuePoolEntry:
    """
    A value pool entry contains the EDIFACT qualifier, a meaning (German text) and an ahb expression.
    A value pool consists of 1 to n ValuePoolEntries.
    The data element 3055 in UTILMD is a good example for a value pool.
    It is used in the segments NAD+MS and NAD+MR. Its ValuePoolEntries are:
    - (key: "9", meaning: "GS1", ahb_expression: "X")
    - (key: "293", meaning: "DE, BDEW", ahb_expression: "X")
    - (key: "332", meaning: "DE, DVGW", ahb_expression: "X")
    """

    #: the qualifier in edifact, might be f.e. "E01", "D", "9", "1.1a", "G_0057"
    qualifier: str = attr.field(validator=attrs.validators.matches_re(r"^[A-Z\d\.a-z_]+$"))
    #: the meaning as it is written in the AHB (f.e. "Einzug", "Entwurfs-Version", "GS1", "Codeliste Gas G_0057"
    meaning: str = attr.field(validator=attrs.validators.instance_of(str))
    #: the ahb expression, in most cases this is a simple "X"; it must not be empty
    ahb_expression: str = attr.field(validator=_check_that_string_is_not_whitespace_or_empty)
    # must not be empty (if so, the value pool entry should not be included of the result)


class ValuePoolEntrySchema(Schema):
    """
    A Schema to serialize ValuePoolEntries
    """

    # this looks like a plain Dict[str,str] but we prefer typed access over loose string key value pairs
    qualifier = fields.String(required=True)
    meaning = fields.String(required=True)
    ahb_expression = fields.String(required=True)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> ValuePoolEntry:
        """
        Converts the barely typed data dictionary into an actual :class:`.ValuePoolEntry`
        """
        return ValuePoolEntry(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class DataElementValuePool(DataElement):
    """
    A DataElementValuePool is a data element with a finite set of allowed values.
    These allowed values are referred to as keys of the "value pool".
    The value pool is defined both on MIG level and AHB level
    The set of values allowed according to the AHB is always a subset of the values allowed according to the MIG.
    """

    value_type: Optional[DataElementDataType] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(DataElementDataType)),  # type:ignore[arg-type]
        default=DataElementDataType.VALUE_POOL,
    )  #: type of the value, if known
    value_pool: List[ValuePoolEntry] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(ValuePoolEntry),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )
    """
    The value pool contains at least one value :class:`.ValuePoolEntry`
    """


class DataElementValuePoolSchema(DataElementSchema):
    """
    A Schema to serialize DataElementValuePool
    """

    value_pool = fields.List(fields.Nested(ValuePoolEntrySchema), required=True)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> DataElementValuePool:
        """
        Converts the barely typed data dictionary into an actual :class:`.DataElementValuePool`
        """
        return DataElementValuePool(**data)


class _FreeTextOrValuePool:
    """
    A class that is easily serializable as dictionary and allows us to _not_ use the marshmallow-union package.
    """

    def __init__(
        self, free_text: Optional[DataElementFreeText] = None, value_pool: Optional[DataElementValuePool] = None
    ):
        self.free_text = free_text
        self.value_pool = value_pool

    free_text: Optional[DataElementFreeText] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(DataElementFreeText))
    )
    value_pool: Optional[DataElementValuePool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(DataElementValuePool))
    )


class _FreeTextOrValuePoolSchema(Schema):
    """
    A schema that represents data of the kind Union[DataElementFreeText,DataElementValuePool]
    There is a python package for that: https://github.com/adamboche/python-marshmallow-union
    but is only has 15 stars; not sure if it's worth the dependency
    """

    # disable unnecessary lambda warning because of circular imports
    free_text = fields.Nested("DataElementFreeTextSchema", allow_none=True, required=False)
    value_pool = fields.Nested("DataElementValuePoolSchema", required=False, allow_none=True)
    # see https://github.com/fuhrysteve/marshmallow-jsonschema/issues/164

    # pylint:disable= unused-argument, no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> Type[DataElement]:
        """
        select the correct part of the data to deserialize
        """
        if "free_text" in data and data["free_text"]:
            return data["free_text"]
        if "value_pool" in data and data["value_pool"]:
            return data["value_pool"]
        return data

    # pylint:disable= unused-argument, no-self-use
    @post_dump
    def post_dump_helper(self, data, **kwargs) -> dict:
        """
        truncate the part _FreeTextOrValuePool that is not needed
        """
        if "value_pool" in data and data["value_pool"]:
            return data["value_pool"]
        if "free_text" in data and data["free_text"]:
            return data["free_text"]
        raise NotImplementedError(f"Data {data} is not implemented for JSON serialization")

    # pylint:disable= unused-argument, no-self-use
    @pre_load
    def pre_load_helper(self, data, **kwargs) -> dict:
        """
        Put the untyped data into a structure that is deserializable as _FreeTextOrValuePool
        """
        if "value_pool" in data:
            return {
                "free_text": None,
                "value_pool": data,
            }
        if "entered_input" in data:
            return {
                "free_text": data,
                "value_pool": None,
            }
        raise NotImplementedError(f"Data {data} is not implemented for JSON deserialization")

    # pylint:disable= unused-argument, no-self-use
    @pre_dump
    def prepare_for_serialization(self, data, **kwargs) -> _FreeTextOrValuePool:
        """
        Detect if data are FreeText or ValuePool data elements
        """
        if isinstance(data, DataElementValuePool):
            return _FreeTextOrValuePool(free_text=None, value_pool=data)
        if isinstance(data, DataElementFreeText):
            return _FreeTextOrValuePool(free_text=data, value_pool=None)
        raise NotImplementedError(f"Data type of {data} is not implemented for JSON serialization")


@attrs.define(auto_attribs=True, kw_only=True)
class SegmentLevel(ABC):
    """
    SegmentLevel describes @annika: what does it describe?
    """

    discriminator: str  # no validator here, because it might be None on initialization and will be set later (trust me)
    ahb_expression: str = attrs.field(validator=attrs.validators.instance_of(str))


class SegmentLevelSchema(Schema):
    """
    A Schema to serialize SegmentLevels
    """

    discriminator = fields.String(required=True)
    ahb_expression = fields.String(required=True)


@attrs.define(auto_attribs=True, kw_only=True)
class Segment(SegmentLevel):
    """
    A Segment contains multiple data elements.
    """

    data_elements: List[DataElement]
    section_name: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )
    """
    For the MIG matching it might be necessary to know the section in which the data element occured in the AHB.
    This might be necessary to e.g. distinguish gas and electricity fields which look the same otherwise.
    See f.e. UTILMD 'Geplante Turnusablesung des MSB (Strom)' vs. 'Geplante Turnusablesung des NB (Gas)'
    """


class SegmentSchema(SegmentLevelSchema):
    """
    A Schema to serialize Segments
    """

    data_elements = fields.List(fields.Nested(_FreeTextOrValuePoolSchema))
    section_name = fields.String(required=False)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> Segment:
        """
        Converts the barely typed data dictionary into an actual :class:`.Segment`
        """
        return Segment(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class SegmentGroup(SegmentLevel):
    """
    A segment group that contains segments and nested groups.
    On root level of a message there might be a "virtual" segment group of all segments that formally don't have a group
    This group has the key "root".
    """

    segments: Optional[List[Segment]] = attrs.field(
        validator=attrs.validators.optional(
            attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(Segment),
                iterable_validator=attrs.validators.instance_of(list),
            )
        ),
        default=None,
    )  #: the segments inside this very group
    segment_groups: Optional[List["SegmentGroup"]] = attrs.field(
        default=None
    )  #: groups that are nested into this group


class SegmentGroupSchema(SegmentLevelSchema):
    """
    A Schema to serialize SegmentGroups.
    """

    segments = fields.List(fields.Nested(SegmentSchema), load_default=None, required=False)
    segment_groups = fields.List(
        fields.Nested("SegmentGroupSchema"),
        # see https://github.com/fuhrysteve/marshmallow-jsonschema/issues/164
        load_default=None,
        required=False,
    )

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> SegmentGroup:
        """
        Converts the barely typed data dictionary into an actual :class:`.SegmentGroup`
        """
        return SegmentGroup(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class EdifactStackLevel:
    """
    The EDIFACT stack level describes the hierarchy level of information inside an EDIFACT message.
    """

    #: the name of the level, f.e. 'Dokument' or 'Nachricht' or 'Meldepunkt'
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    #: describes if this level is groupable / if there are multiple instances of this level within the same message
    is_groupable: bool = attrs.field(validator=attrs.validators.instance_of(bool))
    #: the index if present (f.e. 0)
    index: Optional[int] = attrs.field(
        default=None, validator=attrs.validators.optional(attrs.validators.instance_of(int))
    )


#: a pattern that matches parts of the json path: https://regex101.com/r/iQzdXK/1
_level_pattern = re.compile(r"\[\"(?P<level_name>[^\[\]]+?)\"\](?:\[(?P<index>\d+)\])?")


@attrs.define(auto_attribs=True, kw_only=True)
class EdifactStack:
    """
    The EdifactStack describes where inside an EDIFACT message data are found.
    The stack is independent of the actual implementation used to create the EDIFACT (be it XML, JSON whatever).
    """

    #: levels describe the nesting inside an edifact message
    levels: List[EdifactStackLevel] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(EdifactStackLevel),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )

    @staticmethod
    def from_json_path(json_path: str) -> "EdifactStack":
        """
        reads a json path as it is created by "to_json_path" and returns the corresponding edifact stack
        """
        levels: List[EdifactStackLevel] = []
        for level_match in _level_pattern.finditer(json_path):
            level = EdifactStackLevel(name=level_match["level_name"], is_groupable=level_match["index"] is not None)
            if level.is_groupable:
                level.index = int(level_match["index"])
            levels.append(level)
        return EdifactStack(levels=levels)

    def is_sub_stack_of(self, other: "EdifactStack") -> bool:
        """
        Returns true iff this (self) stack is a sub stack of the other provided stack.
        ([Foo][0][Bar]).is_sub_stack_of([Foo][0]) is true.
        """
        if len(other.levels) > len(self.levels):
            # self cannot be a sub path of other if other is "deeper"
            return False
        for level_self, level_other in zip(other.levels, self.levels):  # , strict=False):  # type:ignore[call-overload]
            # strict is False because it's ok if we stop the iteration if self.levels is "exhausted"; explicit is better
            # the type-ignore for the strict=False is necessary for Python<3.10
            if level_self != level_other:
                return False
        # the iteration stopped meaning that for all levels that both other and self share, they are identical.
        # That's the definition of a sub stack. It also means that any stack is a sub stack of itself.
        return True

    def is_parent_of(self, other: "EdifactStack") -> bool:
        """
        Returns true iff this other stack is a sub stack of self.
        """
        return other.is_sub_stack_of(self)

    def to_json_path(self) -> str:
        """
        Transforms this instance into a JSON Path.
        """
        result: str = "$"
        # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
        # pylint: disable=not-an-iterable
        for level in self.levels:
            result += '["' + level.name + '"]'
            if level.index is not None:
                result += f"[{level.index}]"
            elif level.is_groupable:
                result += "[0]"
        return result


@attrs.define(auto_attribs=True, kw_only=True)
class EdifactStackQuery:
    """
    The EdifactStackQuery contains the data you need to provide to a MIG reader to return you the :class:`EdifactStack`
    of an element
    """

    #: the key of the segment group, f.e. 'root' or 'SG5' or 'SG12'
    segment_group_key: str = attrs.field(validator=attrs.validators.instance_of(str))
    #: the segment code, f.e. 'NAD' or 'DTM'
    segment_code: str = attrs.field(validator=attrs.validators.matches_re("^[A-Z]+$"))
    #: the data element id, f.e. '0068'
    data_element_id: str = attrs.field(validator=attrs.validators.matches_re(r"^\d{4}$"))
    #: the name of the element, f.e. "MP-ID" or "Kundennummer" or "Identifikator"; Is None for Value Pools
    name: Optional[str] = attrs.field(validator=attrs.validators.optional(attrs.validators.instance_of(str)))
    predecessor_qualifier: Optional[str] = attrs.field(
        default=None, validator=attrs.validators.optional(attrs.validators.matches_re(r"^[A-Z\d]+$"))
    )
    """
    Some names are not really unique. F.e. all date time fields carry more or less the same name in the AHB.
    So to distinguish between them you may provide the predecissing qualifier.
    In case of 'DTM+137++what_youre_looking_for' the predecessor qualifier is '137'
    """
    section_name: Optional[str] = attrs.field(
        default=None, validator=attrs.validators.optional(attrs.validators.instance_of(str))
    )
    """
    The section name (f.e. 'Nachrichten-Kopfsegment') might also be used for MIG<->AHB matching
    if the name is too broad or not unique.
    """



"""
Classes that allow to read XML files that contain structural information (Message Implementation Guide information)
"""
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, List, Literal, Optional, TypeVar, Union
from xml.etree.ElementTree import Element

import attrs
from lxml import etree  # type:ignore[import]

from maus import SegmentGroupHierarchy
from maus.models.edifact_components import EdifactStack, EdifactStackLevel, EdifactStackQuery


class MigReader(ABC):
    """
    A MIG reader is a class that reads Message Implementation Guide (MIG) data from a source
    """

    @abstractmethod
    def to_segment_group_hierarchy(self) -> SegmentGroupHierarchy:
        """
        convert the read data into a segment group hierarchy
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")

    @staticmethod
    def are_similar_names(name_x: Optional[str], name_y: Optional[str]) -> bool:
        """
        Returns true if name_x and name_y are somehow similar.
        "Somehow similar" in this context means, that all the artefacts from text to word to PDF + scraping
        + human errors might add up to something which explains the difference between name_x and name_y.
        """
        if (name_x and not name_y) or (name_y and not name_x):
            return False
        if (not name_x) and (not name_y):
            return True
        # neither name_x nor name_y are None below this line
        return MigReader.make_name_comparable(name_x) == MigReader.make_name_comparable(name_y)  # type:ignore[arg-type]

    @staticmethod
    def make_name_comparable(orig_str: str) -> str:
        """
        Removes all the characters that could be a problem when matching names from the AHB with names from the MIG
        """
        result: str = orig_str.lower()
        for removable_character in [" ", "-", "\n"]:
            result = result.replace(removable_character, "")
        return result

    @staticmethod
    def make_tree_names_comparable(tree: etree.ElementTree) -> None:  # pylint:disable=c-extension-no-member
        """
        modifies the provided tree by applying `make_name_comparable` to all name and ahbName attributes
        """
        for element in tree.iter():
            for attrib_key, attrib_value in list(element.attrib.items()):
                if attrib_key in {"name", "ahbName"}:
                    element.attrib[attrib_key] = MigXmlReader.make_name_comparable(attrib_value)

    @staticmethod
    def is_edifact_boilerplate(segment_code: Optional[str]) -> bool:
        """
        returns true iff this segment is not relevant in a sense that it has to be validated or merged with the AHB
        """
        if not segment_code:
            return True
        return segment_code.strip() in {"UNT", "UNZ"}

    # pylint:disable=too-many-arguments
    @abstractmethod
    def get_edifact_stack(self, query: EdifactStackQuery) -> Optional[EdifactStack]:
        """
        Returns the edifact stack for the given combination of segment group, key, data element and name
        """
        raise NotImplementedError("The inheriting class has to implement this method")


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True)
class _MigFilterResult:
    """
    the (internal) result of a query path search inside the tree
    """

    is_unique: Optional[bool]  #: True iff unique, None for no results, False for >1 result
    unique_result: Optional[Element]  #: unique element if there is any; None otherwise
    candidates: Optional[List[Element]]  #: list of candidates if there is >1 result


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True)
class _EdifactStackSearchStrategy:
    """
    The search strategy allows to have a compact yet descriptive representation on how the edifact stack search works.
    The alternative to this is a very nested and hard to understand if/else/then structure with lots of branches.
    Any step inside the strategy has three possible outcomes which are represented by the :class:`_XQueryPathResult`:
    1. There is exactly one unique result => return/exit
    2. There are no results => start over again
    3. There are >1 results => apply additional filters
    """

    #: name, f.e. "filter by data element id"
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    #: the filter is the function that describes the strategy. It consumes the query and (optionally) a list of elements
    filter: Callable[[EdifactStackQuery, Optional[List[Element]]], _MigFilterResult] = attrs.field(
        validator=attrs.validators.is_callable()
    )
    #: The unique result strategy is to return an edifact stack for the unique result element
    unique_result_strategy: Callable[[Element], EdifactStack] = attrs.field(validator=attrs.validators.is_callable())
    #: the no result strategy is to apply another filter based on those candidates that lead to no result (fallback)
    no_result_strategy: Optional["_EdifactStackSearchStrategy"]
    #: in case of multiple results the next strategy uses the multiple results as input (sharpen)
    multiple_results_strategy: Optional["_EdifactStackSearchStrategy"]

    def apply(self, query: EdifactStackQuery, pre_selection: Optional[List[Element]] = None) -> Optional[EdifactStack]:
        """
        Apply the defined strategy until we either have no ideas left or a unique result is found
        """
        # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
        # pylint: disable=not-callable
        filter_result: _MigFilterResult = self.filter(query, pre_selection)
        if filter_result.is_unique is True:
            return self.unique_result_strategy(filter_result.unique_result)  # type:ignore[arg-type]
        if filter_result.candidates and len(filter_result.candidates) > 1:
            if self.multiple_results_strategy is not None:
                return self.multiple_results_strategy.apply(query, filter_result.candidates)
            return None
        if self.no_result_strategy is not None:
            return self.no_result_strategy.apply(query, pre_selection)
        return None


#: a regex to match a ref-segment: https://regex101.com/r/KY25AH/1
_nested_qualifier_pattern = re.compile(r"^(?P<segment_code>[A-Z]+):\d+:\d+\[(?:\w+:)+\w+:?=(?P<qualifier>[A-Z\d]+)\]$")
TResult = TypeVar("TResult")  #: is a type var to indicate an "arbitrary but same" type in a generic function


# pylint:disable=c-extension-no-member
class MigXmlReader(MigReader):
    """
    Reads an XML file an transforms it into a Segment Group Hierarchy
    """

    def __init__(self, init_param: Union[str, Path]):
        self._original_root: etree._Element
        if isinstance(init_param, str):
            self._original_root = etree.fromstring(init_param)
            self._sanitized_root = etree.fromstring(init_param)
        elif isinstance(init_param, Path):
            self._original_root = etree.parse(str(init_param.absolute())).getroot()
            self._sanitized_root = etree.parse(str(init_param.absolute())).getroot()
        else:
            raise ValueError(f"The type of '{init_param}' is not valid")
        # the original tree is the unmodified MIG XML Structure with all its quircks
        self._original_tree: etree.ElementTree = etree.ElementTree(self._original_root)
        # but turns out it's much easier to handle a sanitized tree that is simplified in a sense that
        # * has only lower case (ahb)names which are easy to match because they don't contain whitespace,"-" or casing
        # * has the same structure as the _original_tree so that absolute path expressions from the sanitized tree match
        self._sanitized_tree: etree.ElementTree = etree.ElementTree(self._sanitized_root)
        MigReader.make_tree_names_comparable(self._sanitized_tree)

    def get_format_name(self) -> str:
        """
        the root element of the XML is the name of the EDIFACT format
        """
        return self._original_root.tag

    def element_to_edifact_stack(self, element: etree.Element, use_sanitized_tree: bool) -> EdifactStack:
        """
        extract the edifact seed path from the given element
        :return:
        """
        if use_sanitized_tree:
            xpath = self._sanitized_tree.getpath(element)
        else:
            xpath = self._original_tree.getpath(element)
        stack = EdifactStack(levels=[])
        iter_path = "/" + xpath.split("/")[1]
        for leaf in xpath.split("/")[2:]:
            iter_path += "/" + leaf
            # here we _always_ need to use the original root!
            leaf_element = self._original_root.xpath(iter_path)[0]  # type:ignore[attr-defined]
            level_name: str
            if "ahbName" in leaf_element.attrib:
                level_name = leaf_element.attrib["ahbName"]
            else:
                level_name = leaf_element.attrib["name"]
            # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
            # pylint: disable=no-member
            stack.levels.append(EdifactStackLevel(name=level_name, is_groupable=leaf_element.tag == "class"))
        return stack

    @staticmethod
    def _list_to_mig_filter_result(candidates: List[Element]) -> _MigFilterResult:
        if len(candidates) == 0:
            return _MigFilterResult(candidates=None, is_unique=None, unique_result=None)
            # the == 1 case is handled last
        if len(candidates) > 1:
            return _MigFilterResult(candidates=candidates, is_unique=False, unique_result=None)
        return _MigFilterResult(candidates=None, is_unique=True, unique_result=candidates[0])

    def get_unique_result_by_xpath(self, query_path: str, use_sanitized_tree: bool) -> _MigFilterResult:
        """
        Tries to find an element for the given query path.
        If there's exactly 1 result, it is returned.
        If there are 0 or >1 results a ValueError is raised.
        """
        candidates: List[Element]
        if use_sanitized_tree:
            candidates = list(self._sanitized_root.xpath(query_path))
        else:
            candidates = list(self._original_root.xpath(query_path))
        return MigXmlReader._list_to_mig_filter_result(candidates)

    def get_unique_result_by_segment_group(
        self, candidates: List[Element], query: EdifactStackQuery, use_sanitized_tree: bool
    ) -> _MigFilterResult:
        """
        keep those elements that have the correct segment_group_key
        """
        filtered = [
            e
            for e in candidates
            if self.get_parent_segment_group_key(e, use_sanitized_tree=use_sanitized_tree) == query.segment_group_key
        ]
        return MigXmlReader._list_to_mig_filter_result(filtered)

    # pylint:disable=no-self-use
    def get_unique_result_by_predecessor(self, candidates: List[Element], query: EdifactStackQuery) -> _MigFilterResult:
        """
        Keep those elements that have (in the field) the given predecessor qualifier
        """
        relevant_attribute: Literal["key", "ref"]
        if query.segment_code == "RFF":
            relevant_attribute = "key"
        else:
            relevant_attribute = "ref"
        filtered_by_predecessor = [
            c
            for c in candidates
            if MigXmlReader._get_nested_qualifier(relevant_attribute, c) == query.predecessor_qualifier
        ]  # that's a bit dirty, better parse the ref properly instead of string-matching
        return MigXmlReader._list_to_mig_filter_result(filtered_by_predecessor)

    def get_unique_result_by_parent_predecessor(
        self, candidates: List[Element], query: EdifactStackQuery
    ) -> _MigFilterResult:
        """
        Keep those elements that have (in the parent class) the given predecessor qualifier
        """
        # relevant_parent_attribute: Literal["key", "ref"]
        # if query.segment_code == "RFF":
        #    relevant_parent_attribute = "key"
        # else:
        #    relevant_parent_attribute = "ref"
        filtered_by_predecessor = [
            c
            for c in candidates
            if self.get_parent_predecessor(c, use_sanitized_tree=False) == query.predecessor_qualifier
        ]
        return MigXmlReader._list_to_mig_filter_result(filtered_by_predecessor)

    def get_unique_result_by_parent_segment_group(
        self, candidates: List[Element], query: EdifactStackQuery
    ) -> _MigFilterResult:
        """
        Keep those elements that have (in the parent class) the given segment group key
        """
        filtered_by_segment_group_key = [
            c
            for c in candidates
            if self.get_parent_segment_group_key(c, use_sanitized_tree=False) == query.segment_group_key
        ]
        return MigXmlReader._list_to_mig_filter_result(filtered_by_segment_group_key)

    @staticmethod
    def get_unique_result_by_name(candidates: List[Element], query: EdifactStackQuery):
        """
        returns those elements that have the given name
        """
        filtered_by_names = [
            x
            for x in candidates
            if MigReader.are_similar_names(x.attrib["name"], query.name)
            or ("ahbName" in x.attrib and MigReader.are_similar_names(x.attrib["ahbName"], query.name))
        ]
        return MigXmlReader._list_to_mig_filter_result(filtered_by_names)

    @staticmethod
    def get_unique_result_by_section_name(candidates: List[Element], query: EdifactStackQuery):
        """
        keeps those elements from the candidates whose where the name matches the query section name.
        Does _not_ create a new xpath.
        """
        filtered_by_names = [
            x
            for x in candidates
            if MigReader.are_similar_names(x.attrib["name"], query.section_name)
            # or ("ahbName" in x.attrib and MigReader.are_similar_names(x.attrib["ahbName"], query.name))
        ]
        return MigXmlReader._list_to_mig_filter_result(filtered_by_names)

    @staticmethod
    def _get_segment_group_key_or_none(element: Element) -> Optional[str]:
        """
        returns the segment group of element if present; None otherwise
        """
        if "ref" in element.attrib and element.attrib["ref"].startswith("SG"):
            # the trivial case
            return element.attrib["ref"]
        return None

    @staticmethod
    def _get_nested_qualifier(attrib_key: Literal["ref", "key"], element: Element) -> Optional[str]:
        """
        returns the nested qualifier of an element if present; None otherwise
        """
        if attrib_key in element.attrib:
            match = _nested_qualifier_pattern.match(element.attrib[attrib_key])
            if match:
                return match["qualifier"]
        return None

    def _get_parent_x(
        self, element: Element, evaluator: Callable[[Element], TResult], use_sanitized_tree: bool
    ) -> Optional[TResult]:
        """
        get the 'X' property of the parent where 'X' is the result of the evaluator when applied to an element.
        returns None if not found
        """
        result = evaluator(element)
        if result is not None:
            return result
        if use_sanitized_tree:
            xpath = self._sanitized_tree.getpath(element)
        else:
            xpath = self._original_tree.getpath(element)
        path_parts = list(xpath.split("/"))
        sub_paths: List[str] = []
        for depth in range(2, len(path_parts)):
            sub_path = "/".join(path_parts[0:depth])
            sub_paths.append(sub_path)
        # if xpath was "/foo/bar/asd/xyz", sub_paths is ["/foo", "/foo/bar", "/foo/bar/asd"] now (xyz is not contained!)
        sub_paths.reverse()
        for sub_path in sub_paths:  # loop from inner to root
            leaf_element = self._original_root.xpath(sub_path)[0]  # type:ignore[attr-defined]
            result = evaluator(leaf_element)
            if result is not None:
                return result
        return None

    def get_parent_segment_group_key(self, element: Element, use_sanitized_tree: bool) -> Optional[str]:
        """
        iterate from element towards root and return the first segment group found (the one closes to element).
        returns None if no segment group was found
        """
        return self._get_parent_x(
            element, MigXmlReader._get_segment_group_key_or_none, use_sanitized_tree=use_sanitized_tree
        )

    def get_parent_predecessor(self, element: Element, use_sanitized_tree: bool) -> Optional[str]:
        """
        iterate from element towards root and return the first segment group found (the one closes to element).
        returns None if no segment group was found
        """
        return self._get_parent_x(
            element, lambda c: MigXmlReader._get_nested_qualifier("key", c), use_sanitized_tree=use_sanitized_tree
        )

    def _handle_predecessor_if_present(self, query: EdifactStackQuery) -> Optional[_EdifactStackSearchStrategy]:
        """
        return a strategy for the predecessor if it's present in the query
        """
        if not query.predecessor_qualifier:
            return None
        return _EdifactStackSearchStrategy(
            name="I filter by predecessor",
            filter=lambda q, c: self.get_unique_result_by_predecessor(
                c, q  # type:ignore[arg-type]
            ),
            unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                unique_result, use_sanitized_tree=False
            ),
            no_result_strategy=_EdifactStackSearchStrategy(
                name="J filter by parent predecessor after predecessor lead to no result",
                filter=lambda q, c: self.get_unique_result_by_parent_predecessor(
                    c, q  # type:ignore[arg-type]
                ),
                no_result_strategy=_EdifactStackSearchStrategy(
                    name="M filter by section_name==field name",
                    filter=lambda q, c: MigXmlReader.get_unique_result_by_section_name(c, q),  # type:ignore[arg-type]
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    no_result_strategy=None,
                    multiple_results_strategy=None,
                ),
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                multiple_results_strategy=_EdifactStackSearchStrategy(
                    name="K filter by parents segment group",
                    filter=lambda q, c: self.get_unique_result_by_parent_segment_group(
                        c, q  # type:ignore[arg-type]
                    ),
                    no_result_strategy=None,  # filter by its own predecessor :-(
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    multiple_results_strategy=None,
                ),  # Z18 goes here
            ),
            multiple_results_strategy=_EdifactStackSearchStrategy(
                # doe snot work yet
                name="L filter by parents segment group because parent predecessor is not unique",
                filter=lambda q, c: self.get_unique_result_by_parent_segment_group(c, q),  # type:ignore[arg-type]
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                multiple_results_strategy=None,
                no_result_strategy=None,
            ),
        )

    def _multiple_data_element_matches_handling(
        self, query: EdifactStackQuery
    ) -> Optional[_EdifactStackSearchStrategy]:
        if query.name is not None:
            return _EdifactStackSearchStrategy(
                name="A filter by element name and ahb name",
                filter=lambda q, c: self.get_unique_result_by_name(c, q),  # type:ignore[arg-type]
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                no_result_strategy=_EdifactStackSearchStrategy(
                    name="B filter by parents name because direct name lead to no result",
                    filter=lambda q, _: self.get_unique_result_by_xpath(
                        # pylint:disable=line-too-long
                        f".//class[@name='{MigReader.make_name_comparable(query.name)}']/*[@meta.id='{query.data_element_id}' and starts-with(@ref, '{query.segment_code}')]",  # type:ignore[arg-type]
                        use_sanitized_tree=True,
                    ),
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=True
                    ),
                    no_result_strategy=_EdifactStackSearchStrategy(
                        name="C filter by parents ahb name",
                        filter=lambda q, _: self.get_unique_result_by_xpath(
                            # pylint:disable=line-too-long
                            f".//class[@ahbName='{MigReader.make_name_comparable(query.name)}']/*[@meta.id='{query.data_element_id}' and starts-with(@ref, '{query.segment_code}')]",  # type:ignore[arg-type]
                            use_sanitized_tree=True,
                        ),
                        unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                            unique_result, use_sanitized_tree=True
                        ),
                        multiple_results_strategy=None,
                        no_result_strategy=self._handle_predecessor_if_present(query),
                    ),
                    multiple_results_strategy=None,
                ),
                multiple_results_strategy=_EdifactStackSearchStrategy(
                    name="D filter by segment group",
                    filter=lambda q, c: self.get_unique_result_by_segment_group(
                        c, q, use_sanitized_tree=False  # type:ignore[arg-type]
                    ),
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    multiple_results_strategy=_EdifactStackSearchStrategy(
                        name="E filter by predecessor after segment group filter was not unique",
                        filter=lambda q, c: self.get_unique_result_by_predecessor(
                            c, q  # type:ignore[arg-type]
                        ),
                        unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                            unique_result, use_sanitized_tree=False
                        ),
                        multiple_results_strategy=None,  # raise ValueError("Predecessor B"),
                        no_result_strategy=_EdifactStackSearchStrategy(
                            name="F filter by parent predecessor after predecessor lead to no result",
                            filter=lambda q, c: self.get_unique_result_by_parent_predecessor(
                                c, q  # type:ignore[arg-type]
                            ),
                            no_result_strategy=None,
                            multiple_results_strategy=None,
                            unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                                unique_result, use_sanitized_tree=False
                            ),
                        ),
                    ),
                    no_result_strategy=None,
                ),
            )
        if query.name is None and query.predecessor_qualifier is not None:
            return _EdifactStackSearchStrategy(
                name="G filter by parent predecessor (no fallback)",
                filter=lambda q, c: self.get_unique_result_by_parent_predecessor(
                    c, q  # type:ignore[arg-type]
                ),
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                no_result_strategy=_EdifactStackSearchStrategy(
                    name="H filter by predecessor after parent predecessor lead to no result",
                    filter=lambda q, c: self.get_unique_result_by_predecessor(
                        c, q  # type:ignore[arg-type]
                    ),
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    multiple_results_strategy=None,
                    no_result_strategy=None,
                ),
                multiple_results_strategy=None,
            )
        return None

    def get_edifact_stack(self, query: EdifactStackQuery) -> Optional[EdifactStack]:
        """
        get the edifact stack for the given segment_group, segment... combination or None if there is no match
        """
        strategy = _EdifactStackSearchStrategy(
            name="filter by data element id",
            filter=lambda q, _: (
                self.get_unique_result_by_xpath(
                    f".//*[@meta.id='{q.data_element_id}' and starts-with(@ref, '{q.segment_code}')]",
                    use_sanitized_tree=False,
                )
            ),
            unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                unique_result, use_sanitized_tree=False
            ),
            no_result_strategy=None,  # if there is no matching data element, there's no chance to find a stack
            multiple_results_strategy=self._multiple_data_element_matches_handling(query),
        )
        return strategy.apply(query)

    def to_segment_group_hierarchy(self) -> SegmentGroupHierarchy:
        """
        convert the read data into a segment group hierarchy
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")



from typing import List, Optional, Tuple

import pytest  # type:ignore[import]

from maus.models.message_implementation_guide import SegmentGroupHierarchy, SegmentGroupHierarchySchema
from unittests.serialization_test_helper import assert_serialization_roundtrip  # type:ignore[import]

ALL_SGH_FILES = pytest.mark.datafiles(
    "./unittests/migs/FV2204/segment_group_hierarchies/sgh_mscons.json",
    "./unittests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json",
    "./unittests/migs/FV2204/segment_group_hierarchies/sgh_reqote.json",
    "./unittests/migs/FV2204/segment_group_hierarchies/sgh_iftsta.json",
)


class TestMig:
    """
    Tests the behaviour of the Message Implementation Guide model
    """

    @pytest.mark.parametrize(
        "sgh, expected_json_dict",
        [
            pytest.param(
                # this is basically UTILMD as of FV2204
                SegmentGroupHierarchy(
                    segment_group=None,
                    opening_segment="UNH",
                    sub_hierarchy=[
                        SegmentGroupHierarchy(segment_group="SG1", opening_segment="RFF", sub_hierarchy=[]),
                        SegmentGroupHierarchy(
                            segment_group="SG2",
                            opening_segment="NAD",
                            sub_hierarchy=[
                                SegmentGroupHierarchy(segment_group="SG3", opening_segment="CTA", sub_hierarchy=[])
                            ],
                        ),
                        SegmentGroupHierarchy(
                            segment_group="SG4",
                            opening_segment="IDE",
                            sub_hierarchy=[
                                SegmentGroupHierarchy(
                                    segment_group="SG5",
                                    opening_segment="LOC",
                                    sub_hierarchy=None,
                                ),
                                SegmentGroupHierarchy(segment_group="SG6", opening_segment="RFF", sub_hierarchy=None),
                                SegmentGroupHierarchy(
                                    segment_group="SG8",
                                    opening_segment="SEQ",
                                    sub_hierarchy=[
                                        SegmentGroupHierarchy(
                                            segment_group="SG9", opening_segment="QTY", sub_hierarchy=None
                                        ),
                                        SegmentGroupHierarchy(
                                            segment_group="SG10", opening_segment="CCI", sub_hierarchy=None
                                        ),
                                    ],
                                ),
                                SegmentGroupHierarchy(segment_group="SG12", opening_segment="NAD", sub_hierarchy=None),
                            ],
                        ),
                    ],
                ),
                {
                    "opening_segment": "UNH",
                    "segment_group": None,
                    "sub_hierarchy": [
                        {"opening_segment": "RFF", "segment_group": "SG1", "sub_hierarchy": []},
                        {
                            "opening_segment": "NAD",
                            "segment_group": "SG2",
                            "sub_hierarchy": [{"opening_segment": "CTA", "segment_group": "SG3", "sub_hierarchy": []}],
                        },
                        {
                            "opening_segment": "IDE",
                            "segment_group": "SG4",
                            "sub_hierarchy": [
                                {"opening_segment": "LOC", "segment_group": "SG5", "sub_hierarchy": None},
                                {"opening_segment": "RFF", "segment_group": "SG6", "sub_hierarchy": None},
                                {
                                    "opening_segment": "SEQ",
                                    "segment_group": "SG8",
                                    "sub_hierarchy": [
                                        {"opening_segment": "QTY", "segment_group": "SG9", "sub_hierarchy": None},
                                        {"opening_segment": "CCI", "segment_group": "SG10", "sub_hierarchy": None},
                                    ],
                                },
                                {"opening_segment": "NAD", "segment_group": "SG12", "sub_hierarchy": None},
                            ],
                        },
                    ],
                },
                id="UTILMD",
            ),
        ],
    )
    def test_segment_group_hierarchy_serialization_roundtrip(
        self, sgh: SegmentGroupHierarchy, expected_json_dict: dict
    ):
        assert_serialization_roundtrip(sgh, SegmentGroupHierarchySchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "sgh_file_path, expected_flattened",
        [
            pytest.param(
                "sgh_utilmd.json",
                [
                    (None, "UNH"),
                    ("SG1", "RFF"),
                    ("SG2", "NAD"),
                    ("SG3", "CTA"),
                    ("SG4", "IDE"),
                    ("SG5", "LOC"),
                    ("SG6", "RFF"),
                    ("SG8", "SEQ"),
                    ("SG9", "QTY"),
                    ("SG10", "CCI"),
                    ("SG12", "NAD"),
                ],
            ),
            pytest.param(
                "sgh_mscons.json",
                [
                    (None, "UNB"),
                    ("SG1", "RFF"),
                    ("SG2", "NAD"),
                    ("SG4", "CTA"),
                    ("SG5", "NAD"),
                    ("SG6", "LOC"),
                    ("SG7", "RFF"),
                    ("SG8", "CCI"),
                    ("SG9", "LIN"),
                    ("SG10", "QTY"),
                ],
            ),
            pytest.param(
                "sgh_reqote.json",
                [
                    (None, "UNH"),
                    ("SG1", "RFF"),
                    ("SG11", "NAD"),
                    ("SG14", "CTA"),
                    ("SG27", "LIN"),
                ],
            ),
        ],
    )
    @ALL_SGH_FILES
    def test_flattening(self, datafiles, sgh_file_path: str, expected_flattened: List[Tuple[Optional[str], str]]):
        with open(datafiles / sgh_file_path, "r", encoding="utf-8") as sgh_file:
            sgh = SegmentGroupHierarchySchema().loads(sgh_file.read())
        actual = sgh.flattened()
        assert actual == expected_flattened

    @pytest.mark.parametrize(
        "sgh",
        [
            pytest.param(
                SegmentGroupHierarchy(
                    segment_group="SG4",
                    opening_segment="IDE",
                    sub_hierarchy=[
                        SegmentGroupHierarchy(
                            segment_group="SG5",
                            opening_segment="LOC",
                            sub_hierarchy=None,
                        ),
                        SegmentGroupHierarchy(segment_group="SG6", opening_segment="RFF", sub_hierarchy=None),
                        SegmentGroupHierarchy(
                            segment_group="SG8",
                            opening_segment="SEQ",
                            sub_hierarchy=[
                                SegmentGroupHierarchy(segment_group="SG9", opening_segment="QTY", sub_hierarchy=None),
                                SegmentGroupHierarchy(segment_group="SG10", opening_segment="CCI", sub_hierarchy=None),
                            ],
                        ),
                        SegmentGroupHierarchy(segment_group="SG12", opening_segment="NAD", sub_hierarchy=None),
                    ],
                ),
                id="UTILMD SG4",
            )
        ],
    )
    @pytest.mark.parametrize(
        "sg_key, expected_is_below",
        [
            pytest.param("SG5", True, id="first level below, first entry"),
            pytest.param("SG8", True, id="first level below, third entry"),
            pytest.param("SG9", True, id="second level below, first entry"),
            pytest.param("SG10", True, id="second level below, second entry"),
            pytest.param("SG2", False, id="does not exist (is above)"),
            pytest.param("SG4", False, id="itself is not below"),
            pytest.param(None, False, id="None is above SG4 (not shown)"),
        ],
    )
    def test_is_hierarchically_below(self, sgh: SegmentGroupHierarchy, sg_key: Optional[str], expected_is_below: bool):
        actual = sgh.is_hierarchically_below(sg_key)
        assert actual == expected_is_below



"""
A module to mix/join information from the deep ahb and the MIG (beyond SGH)
"""
from typing import List, Optional

from maus import DataElementFreeText, DataElementValuePool, DeepAnwendungshandbuch
from maus.models.edifact_components import EdifactStackQuery, Segment, SegmentGroup
from maus.reader.mig_reader import MigReader


def _replace_disciminators_with_edifact_stack_segments(
    segments: List[Segment], segment_group_key: str, mig_reader: MigReader, ignore_errors: bool = False
) -> List[Segment]:
    result = segments.copy()
    for segment_index, segment in enumerate(segments):  # type:ignore[union-attr]
        if mig_reader.is_edifact_boilerplate(segment.discriminator):
            continue
        current_segment_key = segment.discriminator
        predecessor_qualifier: Optional[str] = None
        for de_index, data_element in enumerate(segment.data_elements):
            if isinstance(data_element, DataElementFreeText):
                query = EdifactStackQuery(
                    segment_group_key=segment_group_key,
                    segment_code=current_segment_key,
                    data_element_id=data_element.data_element_id,
                    name=data_element.discriminator,
                    predecessor_qualifier=predecessor_qualifier,
                    section_name=segment.section_name,
                )
                stack = mig_reader.get_edifact_stack(query)
                # for easy error analysis set a conditional break point here for 'stack is None'
                if stack is not None:
                    result[segment_index].data_elements[de_index].discriminator = stack.to_json_path()
                else:
                    if not ignore_errors:
                        raise ValueError(f"Couldn't find a stack for {query}")
            if isinstance(data_element, DataElementValuePool):
                query = EdifactStackQuery(
                    segment_group_key=segment_group_key,
                    segment_code=current_segment_key,
                    data_element_id=data_element.data_element_id,
                    name=None,
                    predecessor_qualifier=predecessor_qualifier,
                    section_name=segment.section_name,
                )
                edifact_stack = mig_reader.get_edifact_stack(query)
                if edifact_stack is not None:
                    result[segment_index].data_elements[de_index].discriminator = edifact_stack.to_json_path()
                if edifact_stack is None and len(data_element.value_pool) > 1:
                    if not ignore_errors:
                        raise ValueError(f"Any value pool with more than 1 entry has to have an edifact stack {query}")
                if len(data_element.value_pool) == 1:
                    predecessor_qualifier = data_element.value_pool[0].qualifier
                else:
                    predecessor_qualifier = None
    return result


def _replace_discriminators_with_edifact_stack_groups(
    segment_groups: List[SegmentGroup], mig_reader: MigReader, ignore_errors: bool = False
) -> List[SegmentGroup]:
    """
    replaces all discriminators in the given list of segment groups with an edifact seed path from the provided reader
    """
    result = segment_groups.copy()
    for segment_group_index, segment_group in enumerate(segment_groups):
        current_segment_group_key = segment_group.discriminator
        if result[segment_group_index].segments is not None:
            result[segment_group_index].segments = _replace_disciminators_with_edifact_stack_segments(
                segments=result[segment_group_index].segments,  # type:ignore[arg-type]
                segment_group_key=current_segment_group_key,
                mig_reader=mig_reader,
                ignore_errors=ignore_errors,
            )
        if result[segment_group_index].segment_groups is not None:
            result[segment_group_index].segment_groups = _replace_discriminators_with_edifact_stack_groups(
                segment_groups=segment_group.segment_groups,  # type:ignore[arg-type]
                mig_reader=mig_reader,
                ignore_errors=ignore_errors,
            )
    return result


def replace_discriminators_with_edifact_stack(
    deep_ahb: DeepAnwendungshandbuch, mig_reader: MigReader, ignore_errors: bool = False
) -> None:
    """
    replaces all discriminators in deep_ahb with an edifact seed path from the provided mig_reader
    """
    deep_ahb.lines = _replace_discriminators_with_edifact_stack_groups(deep_ahb.lines, mig_reader, ignore_errors)



# pylint:disable=too-few-public-methods
"""
This module contains classes that are required to model "Anwendungshandbücher" (AHB). There are two kinds of AHBs:
1. the "flat" AHB is very similar to the flat structure scraped from the official PDF files. It has no deep
structure.
2. the "nested" AHB which contains structural information (e.g. that a segment group is contained in
another segment group)
"""

from typing import List, Optional, Sequence
from uuid import UUID

import attrs
from marshmallow import Schema, fields, post_load  # type:ignore[import]

from maus.models.edifact_components import SegmentGroup, SegmentGroupSchema


# pylint:disable=too-many-instance-attributes
@attrs.define(auto_attribs=True, kw_only=True)
class AhbLine:
    """
    An AhbLine is a single line inside the machine readable, flat AHB.
    """

    guid: Optional[UUID] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(UUID))
    )  #: optional key
    # because the combination (segment group, segment, data element, name) is not guaranteed to be unique
    # yes, it's actually that bad already
    segment_group_key: Optional[str] = attrs.field(
        validator=attrs.validators.optional(validator=attrs.validators.instance_of(str))
    )
    """ the segment group, e.g. 'SG5' """

    segment_code: Optional[str] = attrs.field(
        validator=attrs.validators.optional(validator=attrs.validators.instance_of(str))
    )
    """the segment, e.g. 'IDE'"""

    data_element: Optional[str] = attrs.field(
        validator=attrs.validators.optional(validator=attrs.validators.instance_of(str))
    )
    """ the data element ID, e.g. '3224' """

    value_pool_entry: Optional[str] = attrs.field(
        validator=attrs.validators.optional(validator=attrs.validators.instance_of(str))
    )
    """ one of (possible multiple) allowed values, e.g. 'E01' or '293' """

    name: Optional[str] = attrs.field(validator=attrs.validators.optional(validator=attrs.validators.instance_of(str)))
    """the name, e.g. 'Meldepunkt'. It can be both the description of a field but also its meaning"""

    # Check the unittest test_csv_file_reading_11042 to see the different values of name. It's not only the grey fields
    # where user input is expected but also the meanings / values of value pool entries. This means the exact meaning of
    # name can only be determined in the context in which it is used. This is one of many shortcoming of the current AHB
    # structure: Things in the same column don't necessarily mean the same thing.
    ahb_expression: Optional[str] = attrs.field(
        validator=attrs.validators.optional(validator=attrs.validators.instance_of(str))
    )
    """a requirement indicator + an optional condition ("ahb expression"), f.e. 'Muss [123] O [456]' """
    # note: to parse expressions from AHBs consider using AHBicht: https://github.com/Hochfrequenz/ahbicht/

    section_name: Optional[str] = attrs.field(
        validator=attrs.validators.optional(validator=attrs.validators.instance_of(str)), default=None
    )
    """
    The section name describes the purpose of a segment, f.e. "Nachrichten-Kopfsegment" or "Beginn der Nachricht"
    """

    def holds_any_information(self) -> bool:
        """
        Returns true iff the line holds any information exception for just a GUID.
        This is useful to filter out empty lines which are artefacts remaining from the scraping process.
        """
        # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
        # pylint: disable=no-member
        return (
            (self.segment_group_key is not None and len(self.segment_group_key.strip()) > 0)
            or (self.segment_code is not None and len(self.segment_code.strip()) > 0)
            or (self.data_element is not None and len(self.data_element.strip()) > 0)
            or (self.value_pool_entry is not None and len(self.value_pool_entry.strip()) > 0)
            or (self.name is not None and len(self.name.strip()) > 0)
            or (self.ahb_expression is not None and len(self.ahb_expression.strip()) > 0)
        )

    def get_discriminator(self, include_name: bool = True) -> str:
        """
        Generate a unique yet readable discriminator for this given line.
        This discriminator is generated just from the line itself without any information on where it occurs.
        It does neither know its position inside the AHB nor parent or sub groups.
        """
        result: str
        if self.segment_group_key:
            result = f"{self.segment_group_key}->"
        else:
            result = ""
        result += f"{self.segment_code}->{self.data_element}"
        if self.name and include_name:
            result += f" ({self.name})"
        return result


class AhbLineSchema(Schema):
    """
    A schema to (de-)serialize :class:`.AhbLine`
    """

    guid = fields.UUID(required=False, load_default=None)
    segment_group_key = fields.String(required=False, load_default=None)
    segment_code = fields.String(required=False, load_default=None)
    data_element = fields.String(required=False, load_default=None)
    value_pool_entry = fields.String(required=False, load_default=None)
    name = fields.String(required=False, load_default=None)
    ahb_expression = fields.String(required=False, load_default=None)
    section_name = fields.String(required=False, load_default=None)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> AhbLine:
        """
        Converts the barely typed data dictionary into an actual :class:`.AhbLine`
        """
        return AhbLine(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class AhbMetaInformation:
    """
    Meta information about an AHB like f.e. its title, Prüfidentifikator, possible sender and receiver roles
    """

    pruefidentifikator: str  #: identifies the message type (within a fixed format version) e.g. "11042" or "13012"
    # there's more to come  but for now we'll leave it as is, because we're just in a proof of concept phase


class AhbMetaInformationSchema(Schema):
    """
    A schema to (de-)serialize :class:`.AhbMetaInformation`
    """

    pruefidentifikator = fields.String(required=True)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> AhbMetaInformation:
        """
        Converts the barely typed data dictionary into an actual :class:`.AhbMetaInformation`
        """
        return AhbMetaInformation(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class FlatAnwendungshandbuch:
    """
    A flat Anwendungshandbuch (AHB) models an Anwendungshandbuch as combination of some meta data and an ordered list
    of `.class:`.FlatAhbLine s. Basically a flat Anwendungshandbuch is the result of a simple scraping approach.
    You can create instances of this class without knowing anything about the "under the hood" structure of AHBs or MIGs
    """

    meta: AhbMetaInformation = attrs.field(validator=attrs.validators.instance_of(AhbMetaInformation))
    """information about this AHB"""

    lines: List[AhbLine] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(AhbLine),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )  #: ordered list lines as they occur in the AHB

    def get_segment_groups(self) -> List[Optional[str]]:
        """
        :return: a set with all segment groups in this AHB in the order in which they occur
        """
        return FlatAnwendungshandbuch._get_available_segment_groups(self.lines)

    @staticmethod
    def _get_available_segment_groups(lines: List[AhbLine]) -> List[Optional[str]]:
        """
        extracts the distinct segment groups from a list of ahb lines
        :param lines:
        :return: distinct segment groups, including None in the order in which they occur
        """
        # this code is in a static method to make it easily accessible for fine grained unit testing
        result: List[Optional[str]] = []
        for line in lines:
            if line.segment_group_key not in result:
                # an "in" check against a set would be faster but we want to preserve both order and readability
                result.append(line.segment_group_key)
        return result

    def sort_lines_by_segment_groups(self):
        """
        sorts lines by segment groups while preserving the order inside the groups and the order between the groups.
        """
        self.lines = FlatAnwendungshandbuch._sorted_lines_by_segment_groups(self.lines, self.get_segment_groups())

    @staticmethod
    def _sorted_lines_by_segment_groups(ahb_lines: Sequence[AhbLine], sg_order: List[Optional[str]]) -> List[AhbLine]:
        """
        Calls sorted(...) on the provided list and returns a new list.
        Its purpose is, that if a segment group in the AHB (read from top to bottom in the flat ahb/pdf) is interrupted
        by other segment groups, the lines belonging to the same group will be next to each other.
        This is useful to later use a groupby aggregation that only returns one group key per segment group.

        The sort is stable such that the existing order inside the segment groups is maintained.

        Note that this also means, that the order of the return lines no longer the same as in the flat AHB.

        If you provide this function a list:
        [
            (SG2, Foo),
            (SG2, Bar),
            (SG3, ABC),
            (SG3, DEF),
            (SG2, SomethingElse)
        ]
        The result will be:
        [
            (SG2, Foo),
            (SG2, Bar),
            (SG2, SomethingElse)
            (SG3, ABC),
            (SG3, DEF)
        ]

        See the unittests for details.
        """

        # this code is in a static method to make it easily accessible for fine grained unit testing
        result: List[AhbLine] = sorted(ahb_lines, key=lambda x: x.segment_group_key or "")
        result.sort(key=lambda ahb_line: sg_order.index(ahb_line.segment_group_key))
        return result


class FlatAnwendungshandbuchSchema(Schema):
    """
    A schema to (de-)serialize :class:`.FlatAnwendungshandbuch`
    """

    meta = fields.Nested(AhbMetaInformationSchema)
    lines = fields.List(fields.Nested(AhbLineSchema))

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> FlatAnwendungshandbuch:
        """
        Converts the barely typed data dictionary into an actual :class:`.FlatAnwendungshandbuch`
        """
        return FlatAnwendungshandbuch(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class DeepAnwendungshandbuch:
    """
    The data of the AHB nested as described in the MIG.
    """

    meta: AhbMetaInformation = attrs.field(validator=attrs.validators.instance_of(AhbMetaInformation))
    """information about this AHB"""

    lines: List[SegmentGroup] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(SegmentGroup),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )  #: the nested data


class DeepAnwendungshandbuchSchema(Schema):
    """
    A schema to (de-)serialize :class:`.DeepAnwendungshandbuch`
    """

    meta = fields.Nested(AhbMetaInformationSchema)
    lines = fields.List(fields.Nested(SegmentGroupSchema))

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> DeepAnwendungshandbuch:
        """
        Converts the barely typed data dictionary into an actual :class:`.DeepAnwendungshandbuch`
        """
        return DeepAnwendungshandbuch(**data)



"""
A module that reads AHBs from a given CSV file
"""
import csv
import logging
import re
import uuid
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional, Sequence, TextIO, Tuple

from maus.models.anwendungshandbuch import AhbLine, AhbMetaInformation, FlatAnwendungshandbuch

_pruefi_pattern = re.compile(r"^\d{5}$")  #: five digits
_value_pool_entry_pattern = re.compile(r"^[A-Z0-9]{2,}$")
_numeric_value_pool_entry_pattern = re.compile(r"^\d+(?:\.\d+)?[a-z]?$")
_ebd_code_pattern = re.compile(r"^E_\d+$")
_segment_group_pattern = re.compile(r"^SG\d+$")


# pylint:disable=too-few-public-methods
class FlatAhbReader(ABC):
    """
    An AHB Reader reads AHB data from a source and is able to convert them to a :class:`.FlatAnwendungshandbuch`
    """

    @abstractmethod
    def to_flat_ahb(self) -> FlatAnwendungshandbuch:
        """
        convert the read data into a flat anwendungshandbuch
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")


class FlatAhbCsvReader(FlatAhbReader):
    """
    reads csv files and returns AHBs
    """

    def __init__(self, file_path: Path, pruefidentifikator: Optional[str] = None, encoding="utf-8", delimiter=","):
        self.rows: List[AhbLine] = []
        self._logger = logging.getLogger()
        self.current_section_name: Optional[str] = None
        self.pruefidentifikator = pruefidentifikator
        self.delimiter = delimiter
        with open(file_path, "r", encoding=encoding) as infile:
            # current_section_name: Optional[str]
            for row in self.get_raw_rows(infile):
                ahb_line = self.raw_ahb_row_to_ahbline(row)
                if ahb_line is None:
                    continue
                self.rows.append(ahb_line)

    def get_raw_rows(self, file_handle: TextIO) -> List[dict]:
        """
        reads the input file and returns an iterator over raw lines.
        Override this method if your data source is not a CSV file
        """
        reader = csv.DictReader(file_handle, delimiter=self.delimiter)
        if not self.pruefidentifikator:
            self.pruefidentifikator = FlatAhbCsvReader._get_name_of_expression_column(reader.fieldnames)
        if not self.pruefidentifikator:
            raise ValueError("Cannot find column name for ahb expression")
        return list(reader)

    def raw_ahb_row_to_ahbline(self, ahb_row: dict) -> Optional[AhbLine]:
        """
        Converts a row of the raw/scraped AHB into the AhbLine data structure.
        Returns None for rows that are skipped.
        Override this method is your column names/input data structure differs.
        """
        value_pool_entry, description = FlatAhbCsvReader.separate_value_pool_entry_and_name(
            ahb_row["Codes und Qualifier"], ahb_row["Beschreibung"]
        )
        segment_group: Optional[str] = None
        if FlatAhbCsvReader._is_segment_group(ahb_row["Segment Gruppe"]):
            segment_group = ahb_row["Segment Gruppe"]
        elif len(ahb_row["Segment Gruppe"]) >= 3:
            self.current_section_name = ahb_row["Segment Gruppe"].strip() or None  # f.e. "Nachrichten-Kopfsegment"
            self._logger.debug("Processing %s section '%s'", self.pruefidentifikator, self.current_section_name)
            return None  # possibly a section heading like "Nachrichten-Endesegment"
            # this is different from segment group = None which is value for e.g. the UNH
        result: AhbLine = AhbLine(
            guid=uuid.uuid4(),
            segment_group_key=segment_group,
            segment_code=ahb_row["Segment"] or None,
            data_element=ahb_row["Datenelement"] or None,
            value_pool_entry=value_pool_entry,
            ahb_expression=ahb_row[self.pruefidentifikator] or None,
            name=description,
            section_name=self.current_section_name,
        )
        return result

    @staticmethod
    def separate_value_pool_entry_and_name(
        x: Optional[str], y: Optional[str]  # pylint:disable=invalid-name
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        The PDFs are so broken, that sometimes the Codes column contains the description of the line instead of the code
        This function returns a value_pool_entry at index 0, a possible description at index 1, even if they're mixed up
        in the source files.
        """
        if x is not None:
            x = x.strip()
        if y is not None:
            y = y.strip()
        if FlatAhbCsvReader._is_value_pool_entry(x) and not FlatAhbCsvReader._is_value_pool_entry(y):
            return x, y or None
        if FlatAhbCsvReader._is_value_pool_entry(x) and FlatAhbCsvReader._is_value_pool_entry(y):
            # Both look like a value pool entry. This typically happens f.e. for date qualifiers or code lists
            return x, y
        return y or None, x or None

    @staticmethod
    def _is_value_pool_entry(candidate: Optional[str]) -> bool:
        """
        Returns true iff the provided candidate is a possible value pool entry.
        """
        if not candidate:
            return False
        if _value_pool_entry_pattern.match(candidate) is not None:
            return True
        # numbers alone might be value pool entries even if they don't match the regex
        # we don't use "isdigit" because isdigit f.e. does not match '1.2'
        if _numeric_value_pool_entry_pattern.match(candidate) is not None:
            return True
        if len(candidate) == 1 and candidate.upper() == candidate:
            return True
        return _ebd_code_pattern.match(candidate) is not None

    @staticmethod
    def _is_segment_group(candidate: Optional[str]) -> bool:
        """
        Returns true iff the provided candidate is a possible segment group
        """
        if not candidate:
            return False
        return _value_pool_entry_pattern.match(candidate) is not None

    @staticmethod
    def _get_name_of_expression_column(field_names: Optional[Sequence[str]]) -> Optional[str]:
        """
        Gets the name of the column that holds the AHB expressions.
        This allows us to read any AHB without a priori knowledge of its pruefidentifikator.

        :param field_names: list of fieldnames
        :return: the first field name that is a 5 digit name or None if none is found.
        """
        if not field_names:
            return None
        for field_name in field_names:
            if _pruefi_pattern.match(field_name):
                return field_name
        return None

    def to_flat_ahb(self) -> FlatAnwendungshandbuch:
        """
        Converts the content of the CSV file to a FlatAnwendungshandbuch.
        :return:
        """
        return FlatAnwendungshandbuch(
            meta=AhbMetaInformation(
                pruefidentifikator=self.pruefidentifikator  # type:ignore[arg-type]
            ),
            lines=[row for row in self.rows if row.holds_any_information()],
        )



from pathlib import Path
from typing import List, Optional

import pytest  # type:ignore[import]

from maus.reader.flat_ahb_reader import FlatAhbCsvReader


class TestAhbCsvReader:
    """
    Tests the ahb_csv_reader
    """

    @pytest.mark.parametrize(
        "field_names, expected_column_name",
        [
            pytest.param(
                [
                    "Segment Gruppe",
                    "Segment",
                    "Datenelement",
                    "Codes und Qualifier",
                    "Beschreibung",
                    "11042",
                    "Bedingung",
                ],
                "11042",
            ),
            pytest.param(["foo", "bar"], None),
            pytest.param(None, None),
            pytest.param([], None),
        ],
    )
    def test_ahb_expression_column_finder(self, field_names: List[str], expected_column_name: Optional[str]):
        actual = FlatAhbCsvReader._get_name_of_expression_column(field_names)
        assert actual == expected_column_name

    @pytest.mark.parametrize(
        "value, expected_is_value_pool_entry",
        [
            pytest.param(None, False),
            pytest.param("", False),
            pytest.param("E01", True),
        ],
    )
    def test_is_value_pool_entry(self, value: Optional[str], expected_is_value_pool_entry: bool):
        actual = FlatAhbCsvReader._is_value_pool_entry(value)
        assert actual == expected_is_value_pool_entry

    @pytest.mark.parametrize(
        "value, expected_is_segment_group",
        [
            pytest.param(None, False),
            pytest.param("", False),
            pytest.param("Nachrichten-Endesegment", False),
            pytest.param("SG4", True),
            pytest.param("SG12", True),
        ],
    )
    def test_is_segment_group(self, value: Optional[str], expected_is_segment_group: bool):
        actual = FlatAhbCsvReader._is_segment_group(value)
        assert actual == expected_is_segment_group

    @pytest.mark.parametrize(
        "csv_code, csv_beschreibung, expected_code, expected_beschreibung",
        [
            pytest.param("Nachrichten-Referenznummer", None, None, "Nachrichten-Referenznummer"),
            pytest.param("137", "Dokumenten-/Nachrichtendatum/-zeit", "137", "Dokumenten-/Nachrichtendatum/-zeit"),
            pytest.param(
                "Datum oder Uhrzeit oderZeitspanne, Wert", None, None, "Datum oder Uhrzeit oderZeitspanne, Wert"
            ),
            pytest.param("303", "CCYYMMDDHHMMZZZ", "303", "CCYYMMDDHHMMZZZ"),
            pytest.param("9", "GS1", "9", "GS1"),
            pytest.param("IC", "Informationskontakt", "IC", "Informationskontakt"),
            pytest.param(
                "MS",
                "Dokumenten-/Nachrichtenausstellerbzw. -absender",
                "MS",
                "Dokumenten-/Nachrichtenausstellerbzw. -absender",
            ),
            pytest.param("TN", "Transaktions-Referenznummer", "TN", "Transaktions-Referenznummer"),
            pytest.param(
                "293",
                "DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)",
                "293",
                "DE, BDEW (Bundesverband der Energie- und Wasserwirtschaft e.V.)",
            ),
            pytest.param(
                "1.2",
                "Versionsnummer derzugrundeliegendenBDEW-Nachrichtenbeschreibung",
                "1.2",
                "Versionsnummer derzugrundeliegendenBDEW-Nachrichtenbeschreibung",
                id="format version reqote",
            ),
            pytest.param(
                "D",
                "Entwurfs-Version",
                "D",
                "Entwurfs-Version",
                id="entwurfs version reqote",
            ),
            pytest.param(
                "Dokumenten-/Nachrichtenausstellerbzw. -absender",
                "MS ",  # <-- has a trailing white space
                "MS",  # <-- has no trailing whitespace
                "Dokumenten-/Nachrichtenausstellerbzw. -absender",
                id="strip inputs; utilmd absender",
            ),
            pytest.param("EBD Nr. E_0456", "E_0456", "E_0456", "EBD Nr. E_0456", id="EBD Name in IFSTA"),
            pytest.param(
                "2.0d",
                "Versionsnummer derzugrundeliegendenBDEW-Nachrichtenbeschreibung",
                "2.0d",
                "Versionsnummer derzugrundeliegendenBDEW-Nachrichtenbeschreibung",
                id="iftsta version",
            ),
        ],
    )
    def test_code_description_separation(
        self,
        csv_code: Optional[str],
        csv_beschreibung: Optional[str],
        expected_code: Optional[str],
        expected_beschreibung: Optional[str],
    ):
        actual_code, actual_beschreibung = FlatAhbCsvReader.separate_value_pool_entry_and_name(
            csv_code, csv_beschreibung
        )
        assert actual_code == expected_code
        assert actual_beschreibung == expected_beschreibung

    @pytest.mark.datafiles("./unittests/ahbs/FV2204/UTILMD/11042.csv")
    def test_csv_file_reading_11042(self, datafiles):
        path_to_csv: Path = datafiles / "11042.csv"
        reader = FlatAhbCsvReader(file_path=path_to_csv)
        assert len(reader.rows) == 846
        # first row assertions
        first_row = reader.rows[0]
        assert first_row.segment_code == "UNH"
        assert first_row.ahb_expression == "Muss"

        # last row assertions
        last_row = reader.rows[845]
        assert last_row.segment_code == "UNT"
        assert last_row.data_element == "0062"
        assert last_row.name == "Nachrichten-Referenznummer"
        assert last_row.ahb_expression == "X"

        flat_ahb = reader.to_flat_ahb()
        assert len(flat_ahb.lines) < len(reader.rows)  # filter out the empty lines
        assert flat_ahb.get_segment_groups() == [None, "SG2", "SG3", "SG4", "SG5", "SG6", "SG8", "SG9", "SG10", "SG12"]



"""
MAUS is the MIG AHB Utility stack.
This module contains methods to merge data from Message Implementation Guide and Anwendungshandbuch
"""
from itertools import groupby
from typing import List, Sequence

from maus.models.anwendungshandbuch import AhbLine, DeepAnwendungshandbuch, FlatAnwendungshandbuch
from maus.models.edifact_components import (
    DataElement,
    DataElementFreeText,
    DataElementValuePool,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
    derive_data_type_from_segment_code,
)
from maus.models.message_implementation_guide import SegmentGroupHierarchy


def merge_lines_with_same_data_element(ahb_lines: Sequence[AhbLine]) -> DataElement:
    """
    Merges lines that have the same data element into a single data element instance which is returned
    """
    distinct_data_element_keys = {ahb_line.data_element for ahb_line in ahb_lines}
    if len(distinct_data_element_keys) != 1:
        raise ValueError(
            "You must only use this function with lines that share the same data element but the "
            f"parameter ahb_lines contains: {', '.join([x or '' for x in distinct_data_element_keys])} "
        )
    result: DataElement
    if ahb_lines[0].value_pool_entry is not None:
        result = DataElementValuePool(
            discriminator=ahb_lines[0].get_discriminator(include_name=False),
            value_pool=[],
            data_element_id=ahb_lines[0].data_element,  # type:ignore[arg-type]
        )
        for data_element_value_entry in ahb_lines:
            if data_element_value_entry.name is None:
                # pylint:disable=fixme
                # todo: this is line where there is only a description and nothing else. i hate it
                # f.e. there is 35001: https://github.com/Hochfrequenz/mig_ahb_utility_stack/issues/21
                continue
            if not data_element_value_entry.ahb_expression:
                # value pool entries with empty/None AHB expression shall not be included
                # https://github.com/Hochfrequenz/mig_ahb_utility_stack/issues/38
                continue
            value_pool_entry = ValuePoolEntry(
                qualifier=data_element_value_entry.value_pool_entry,  # type:ignore[arg-type]
                meaning=data_element_value_entry.name.strip(),  # type:ignore[assignment,union-attr]
                ahb_expression=data_element_value_entry.ahb_expression,
            )
            result.value_pool.append(value_pool_entry)  # type:ignore[index]
    else:
        result = DataElementFreeText(
            entered_input=None,
            ahb_expression=ahb_lines[0].ahb_expression or "",
            discriminator=ahb_lines[0].name or ahb_lines[0].get_discriminator(include_name=True),
            data_element_id=ahb_lines[0].data_element,  # type:ignore[arg-type]
        )
        # a free text field never spans more than 1 line

        data_type = derive_data_type_from_segment_code(ahb_lines[0].segment_code)  # type:ignore[arg-type]
        if data_type is not None:
            result.value_type = data_type
    return result


def group_lines_by_segment(segment_group_lines: List[AhbLine]) -> List[Segment]:
    """
    convert the given lines (which are assumed to be from the same segment group) into single segments
    """
    result: List[Segment] = []
    for segment_key, segments in groupby(segment_group_lines, key=lambda line: line.segment_code):
        if segment_key is None:
            continue  # filter out segment group level
        this_segment_lines = list(segments)
        if not this_segment_lines[0].ahb_expression:
            # segments with an empty AHB expression shall not be included
            # https://github.com/Hochfrequenz/mig_ahb_utility_stack/issues/38
            continue
        segment = Segment(
            discriminator=segment_key,  # type:ignore[arg-type] # shall not be none after sanitizing
            data_elements=[],
            ahb_expression=this_segment_lines[0].ahb_expression,
            section_name=this_segment_lines[0].section_name,
        )
        for data_element_key, data_element_lines in groupby(this_segment_lines, key=lambda line: line.data_element):
            if data_element_key is None:
                continue
            data_element = merge_lines_with_same_data_element(list(data_element_lines))
            segment.data_elements.append(data_element)  # type:ignore[union-attr] # yes, it's not none
        result.append(segment)
    return result


def group_lines_by_segment_group(
    ahb_lines: List[AhbLine], segment_group_hierarchy: SegmentGroupHierarchy
) -> List[SegmentGroup]:
    """
    Group the lines by their segment group and arrange the segment groups in a flat/not deep list
    """
    result: List[SegmentGroup] = []
    for hierarchy_segment_group, _ in segment_group_hierarchy.flattened():  # flatten = ignore hierarchy, preserve order
        # here we assume, that the ahb_lines are easily groupable
        # (meaning, the ran through FlatAhb.sort_lines_by_segment_groups once)
        for segment_group_key, sg_group in groupby(ahb_lines, key=lambda line: line.segment_group_key):
            if hierarchy_segment_group == segment_group_key:
                this_sg = list(sg_group)
                sg_draft = SegmentGroup(
                    discriminator=segment_group_key,  # type:ignore[arg-type] # might be None now, will be replace later
                    ahb_expression=this_sg[0].ahb_expression or "",  # segment groups barely ever have an expression
                    segments=group_lines_by_segment(this_sg),
                    segment_groups=[],
                )
                result.append(sg_draft)
    return result


def nest_segment_groups_into_each_other(
    flat_groups: List[SegmentGroup], segment_group_hierarchy: SegmentGroupHierarchy
) -> List[SegmentGroup]:
    """
    Take the pre-grouped but flat/"not nested" groups and nest them into each other as described in the SGH
    """
    result: List[SegmentGroup] = []
    for n, segment_group in enumerate(flat_groups):  # pylint:disable=invalid-name
        if segment_group.discriminator == segment_group_hierarchy.segment_group:
            # this is the root level for the given hierarchy
            result.append(segment_group)
            if segment_group_hierarchy.sub_hierarchy is None:
                break
            for sub_sgh in segment_group_hierarchy.sub_hierarchy:
                # pass the remaining groups to to the remaining hierarchy
                sub_results = nest_segment_groups_into_each_other(flat_groups[n + 1 :], sub_sgh)
                for sub_result in sub_results:
                    # for any entry coming from nest_segment_groups_into_each other, it is ensured,
                    # that the segment groups are maybe an empty list but never None.
                    result[-1].segment_groups.append(sub_result)  # type:ignore[union-attr]
    return result


def to_deep_ahb(
    flat_ahb: FlatAnwendungshandbuch, segment_group_hierarchy: SegmentGroupHierarchy
) -> DeepAnwendungshandbuch:
    """
    Converts a flat ahb into a nested ahb using the provided segment hierarchy
    """
    result = DeepAnwendungshandbuch(meta=flat_ahb.meta, lines=[])
    flat_ahb.sort_lines_by_segment_groups()
    flat_groups = group_lines_by_segment_group(flat_ahb.lines, segment_group_hierarchy)
    # in a first step we group the lines by their segment groups but ignore the actual hierarchy except for the order
    result.lines = nest_segment_groups_into_each_other(flat_groups, segment_group_hierarchy)
    for segment_group in result.lines:
        if not isinstance(segment_group, SegmentGroup):
            continue
        if segment_group.discriminator is None:
            segment_group.discriminator = "root"
    return result



import uuid
from typing import List, Optional, Set

import pytest  # type:ignore[import]

from maus.models.anwendungshandbuch import (
    AhbLine,
    AhbLineSchema,
    AhbMetaInformation,
    AhbMetaInformationSchema,
    DeepAnwendungshandbuch,
    DeepAnwendungshandbuchSchema,
    FlatAnwendungshandbuch,
    FlatAnwendungshandbuchSchema,
)
from maus.models.edifact_components import (
    DataElementFreeText,
    DataElementValuePool,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
)
from unittests.serialization_test_helper import assert_serialization_roundtrip  # type:ignore[import]

meta_x = AhbMetaInformation(
    pruefidentifikator="11042",
)
meta_y = AhbMetaInformation(
    pruefidentifikator="11043",
)

line_x = AhbLine(
    ahb_expression="Muss [1] O [2]",
    segment_group_key="SG2",
    segment_code="NAD",
    data_element="3039",
    value_pool_entry="E01",
    name="MP-ID",
    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
)
line_y = AhbLine(
    ahb_expression="Muss [3] O [4]",
    segment_group_key="SG2",
    segment_code="NAD",
    data_element="3039",
    value_pool_entry="E01",
    name="MP-ID",
    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
)


class TestAhb:
    """
    Tests the behaviour of the Anwendungshandbuch model
    """

    @pytest.mark.parametrize(
        "ahb, expected_json_dict",
        [
            pytest.param(
                AhbMetaInformation(
                    pruefidentifikator="11042",
                ),
                {
                    "pruefidentifikator": "11042",
                },
            ),
        ],
    )
    def test_ahb_meta_information_serialization_roundtrip(self, ahb: AhbMetaInformation, expected_json_dict: dict):
        assert_serialization_roundtrip(ahb, AhbMetaInformationSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "ahb_x, ahb_y, are_equal",
        [
            pytest.param(meta_x, AhbMetaInformationSchema().loads(AhbMetaInformationSchema().dumps(meta_x)), True),
            pytest.param(meta_x, meta_y, False),
        ],
    )
    def test_ahb_meta_information_equality(self, ahb_x: AhbMetaInformation, ahb_y: AhbMetaInformation, are_equal: bool):
        actual = ahb_x == ahb_y
        assert actual == are_equal

    @pytest.mark.parametrize(
        "ahb_line, expected_json_dict",
        [
            pytest.param(
                AhbLine(
                    ahb_expression="Muss [1] O [2]",
                    segment_group_key="SG2",
                    segment_code="NAD",
                    data_element="3039",
                    value_pool_entry="E01",
                    name="MP-ID",
                    section_name="Foo",
                    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                ),
                {
                    "ahb_expression": "Muss [1] O [2]",
                    "segment_group_key": "SG2",
                    "segment_code": "NAD",
                    "data_element": "3039",
                    "value_pool_entry": "E01",
                    "name": "MP-ID",
                    "guid": "12b1a98a-edf5-4177-89e5-a6d8a92c5fdc",
                    "section_name": "Foo",
                },
            ),
        ],
    )
    def test_ahbline_serialization_roundtrip(self, ahb_line: AhbLine, expected_json_dict: dict):
        assert_serialization_roundtrip(ahb_line, AhbLineSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "line_x, line_y, are_equal",
        [
            pytest.param(line_x, AhbLineSchema().loads(AhbLineSchema().dumps(line_x)), True),
            pytest.param(line_x, line_y, False),
        ],
    )
    def test_ahb_line_equality(self, line_x: AhbLine, line_y: AhbLine, are_equal: bool):
        actual = line_x == line_y
        assert actual == are_equal

    @pytest.mark.parametrize(
        "ahb_line, expected_holds_any_information",
        [
            pytest.param(
                AhbLine(
                    ahb_expression="Muss [1] O [2]",
                    segment_group_key="SG2",
                    segment_code="NAD",
                    data_element="3039",
                    value_pool_entry="E01",
                    name="MP-ID",
                    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                ),
                True,
            ),
            pytest.param(
                AhbLine(
                    ahb_expression=None,
                    segment_group_key=None,
                    segment_code=None,
                    data_element=None,
                    value_pool_entry=None,
                    name=None,
                    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                ),
                False,
            ),
            pytest.param(
                AhbLine(
                    ahb_expression=" ",
                    segment_group_key="",
                    segment_code="",
                    data_element="   ",
                    value_pool_entry="",
                    name="",
                    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                ),
                False,
            ),
        ],
    )
    def test_ahbline_holds_any_information(self, ahb_line: AhbLine, expected_holds_any_information: bool):
        actual = ahb_line.holds_any_information()
        assert actual == expected_holds_any_information

    @pytest.mark.parametrize(
        "ahb_line, include_name, expected_discriminator",
        [
            pytest.param(
                AhbLine(
                    ahb_expression="Muss [1] O [2]",
                    segment_group_key="SG2",
                    segment_code="NAD",
                    data_element="3039",
                    value_pool_entry="E01",
                    name="MP-ID",
                    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                ),
                True,
                "SG2->NAD->3039 (MP-ID)",
            ),
            pytest.param(
                AhbLine(
                    ahb_expression="Muss [1] O [2]",
                    segment_group_key="SG2",
                    segment_code="NAD",
                    data_element="3039",
                    value_pool_entry="E01",
                    name="MP-ID",
                    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                ),
                False,
                "SG2->NAD->3039",
            ),
            pytest.param(
                AhbLine(
                    ahb_expression="Muss [1] O [2]",
                    segment_group_key=None,
                    segment_code="UNH",
                    data_element="0062",
                    guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                    value_pool_entry=None,
                    name=None,
                ),
                True,
                "UNH->0062",
            ),
        ],
    )
    def test_ahbline_get_discriminator(self, ahb_line: AhbLine, include_name: bool, expected_discriminator: str):
        actual = ahb_line.get_discriminator(include_name)
        assert actual == expected_discriminator

    @pytest.mark.parametrize(
        "flat_ahb, expected_json_dict",
        [
            pytest.param(
                FlatAnwendungshandbuch(
                    meta=AhbMetaInformation(pruefidentifikator="11042"),
                    lines=[
                        AhbLine(
                            ahb_expression="Muss [1] O [2]",
                            segment_group_key="SG2",
                            segment_code="NAD",
                            data_element="3039",
                            value_pool_entry="E01",
                            name="MP-ID",
                            guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                            section_name="MP-ID Absender",
                        )
                    ],
                ),
                {
                    "meta": {"pruefidentifikator": "11042"},
                    "lines": [
                        {
                            "ahb_expression": "Muss [1] O [2]",
                            "segment_group_key": "SG2",
                            "segment_code": "NAD",
                            "data_element": "3039",
                            "value_pool_entry": "E01",
                            "name": "MP-ID",
                            "guid": "12b1a98a-edf5-4177-89e5-a6d8a92c5fdc",
                            "section_name": "MP-ID Absender",
                        }
                    ],
                },
            ),
        ],
    )
    def test_flatahb_serialization_roundtrip(self, flat_ahb: FlatAnwendungshandbuch, expected_json_dict: dict):
        assert_serialization_roundtrip(flat_ahb, FlatAnwendungshandbuchSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "ahb_x, ahb_y, are_equal",
        [
            pytest.param(
                FlatAnwendungshandbuch(
                    meta=meta_x,
                    lines=[line_x],
                ),
                FlatAnwendungshandbuch(
                    meta=meta_x,
                    lines=[line_x],
                ),
                True,
            ),
            pytest.param(
                FlatAnwendungshandbuch(
                    meta=meta_x,
                    lines=[line_x],
                ),
                FlatAnwendungshandbuch(
                    meta=meta_y,
                    lines=[line_x],
                ),
                False,
                id="different meta",
            ),
            pytest.param(
                FlatAnwendungshandbuch(
                    meta=meta_x,
                    lines=[line_x],
                ),
                FlatAnwendungshandbuch(
                    meta=meta_x,
                    lines=[line_y],
                ),
                False,
                id="different line",
            ),
        ],
    )
    def test_flat_ahb_equality(self, ahb_x: FlatAnwendungshandbuch, ahb_y: FlatAnwendungshandbuch, are_equal: bool):
        actual = ahb_x == ahb_y
        assert actual == are_equal

    @pytest.mark.parametrize(
        "deep_ahb, expected_json_dict",
        [
            pytest.param(
                DeepAnwendungshandbuch(
                    meta=AhbMetaInformation(pruefidentifikator="11042"),
                    lines=[
                        SegmentGroup(
                            ahb_expression="expr A",
                            discriminator="disc A",
                            segments=[
                                Segment(
                                    ahb_expression="expr B",
                                    discriminator="disc B",
                                    section_name="foo",
                                    data_elements=[
                                        DataElementValuePool(
                                            value_pool=[
                                                ValuePoolEntry(qualifier="hello", meaning="world", ahb_expression="X"),
                                                ValuePoolEntry(qualifier="maus", meaning="rocks", ahb_expression="X"),
                                            ],
                                            discriminator="baz",
                                            data_element_id="0123",
                                        ),
                                        DataElementFreeText(
                                            ahb_expression="Muss [1]",
                                            entered_input="Hello Maus",
                                            discriminator="bar",
                                            data_element_id="4567",
                                        ),
                                    ],
                                ),
                            ],
                            segment_groups=[
                                SegmentGroup(
                                    discriminator="disc C",
                                    ahb_expression="expr C",
                                    segments=[
                                        Segment(
                                            section_name="bar",
                                            ahb_expression="expr Y",
                                            discriminator="disc Y",
                                            data_elements=[],
                                        )
                                    ],
                                    segment_groups=None,
                                ),
                            ],
                        ),
                    ],
                ),
                {
                    "meta": {"pruefidentifikator": "11042"},
                    "lines": [
                        {
                            "ahb_expression": "expr A",
                            "discriminator": "disc A",
                            "segments": [
                                {
                                    "section_name": "foo",
                                    "ahb_expression": "expr B",
                                    "discriminator": "disc B",
                                    "data_elements": [
                                        {
                                            "value_pool": [
                                                {"qualifier": "hello", "meaning": "world", "ahb_expression": "X"},
                                                {"qualifier": "maus", "meaning": "rocks", "ahb_expression": "X"},
                                            ],
                                            "discriminator": "baz",
                                            "data_element_id": "0123",
                                            "value_type": "VALUE_POOL",
                                        },
                                        {
                                            "ahb_expression": "Muss [1]",
                                            "entered_input": "Hello Maus",
                                            "discriminator": "bar",
                                            "data_element_id": "4567",
                                            "value_type": "TEXT",
                                        },
                                    ],
                                }
                            ],
                            "segment_groups": [
                                {
                                    "ahb_expression": "expr C",
                                    "discriminator": "disc C",
                                    "segments": [
                                        {
                                            "section_name": "bar",
                                            "ahb_expression": "expr Y",
                                            "discriminator": "disc Y",
                                            "data_elements": [],
                                        }
                                    ],
                                    "segment_groups": None,
                                }
                            ],
                        },
                    ],
                },
            ),
        ],
    )
    def test_deep_ahb_serialization_roundtrip(self, deep_ahb: DeepAnwendungshandbuch, expected_json_dict: dict):
        assert_serialization_roundtrip(deep_ahb, DeepAnwendungshandbuchSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "ahb_x, ahb_y, are_equal",
        [
            pytest.param(
                DeepAnwendungshandbuch(
                    meta=AhbMetaInformation(pruefidentifikator="11042"),
                    lines=[
                        SegmentGroup(
                            ahb_expression="expr A",
                            discriminator="disc A",
                            segments=[
                                Segment(
                                    ahb_expression="expr B",
                                    discriminator="disc B",
                                    data_elements=[
                                        DataElementValuePool(
                                            value_pool=[
                                                ValuePoolEntry(qualifier="hello", meaning="world", ahb_expression="X"),
                                                ValuePoolEntry(qualifier="maus", meaning="rocks", ahb_expression="X"),
                                            ],
                                            discriminator="baz",
                                            data_element_id="0123",
                                        ),
                                        DataElementFreeText(
                                            ahb_expression="Muss [1]",
                                            entered_input="Hello Maus",
                                            discriminator="bar",
                                            data_element_id="4567",
                                        ),
                                    ],
                                ),
                            ],
                            segment_groups=[],
                        ),
                    ],
                ),
                DeepAnwendungshandbuch(
                    meta=AhbMetaInformation(pruefidentifikator="11042"),
                    lines=[
                        SegmentGroup(
                            ahb_expression="expr A",
                            discriminator="disc A",
                            segments=[
                                Segment(
                                    ahb_expression="expr B",
                                    discriminator="disc B",
                                    data_elements=[
                                        DataElementValuePool(
                                            value_pool=[
                                                ValuePoolEntry(qualifier="hello", meaning="world", ahb_expression="X"),
                                                ValuePoolEntry(qualifier="maus", meaning="rocks", ahb_expression="X"),
                                            ],
                                            discriminator="baz",
                                            data_element_id="0123",
                                        ),
                                        DataElementFreeText(
                                            ahb_expression="Muss [1]",
                                            entered_input="Hello Mice",
                                            discriminator="bar",
                                            data_element_id="4567",
                                        ),
                                    ],
                                ),
                            ],
                            segment_groups=[],
                        ),
                    ],
                ),
                False,
                id="different lines",
            ),
        ],
    )
    def test_deep_ahb_equality(self, ahb_x: DeepAnwendungshandbuch, ahb_y: DeepAnwendungshandbuch, are_equal: bool):
        actual = ahb_x == ahb_y
        assert actual == are_equal

    @pytest.mark.parametrize(
        "lines, expected_group_names",
        [
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="Foo",
                        guid=None,
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Bar",
                        guid=None,
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key=None,
                        guid=None,
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Bar",
                        guid=None,
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                ],
                ["Foo", "Bar", None],
            )
        ],
    )
    def test_get_segment_groups(self, lines: List[AhbLine], expected_group_names: Set[Optional[str]]):
        actual = FlatAnwendungshandbuch._get_available_segment_groups(lines)
        assert actual == expected_group_names

    @pytest.mark.parametrize(
        "unsorted_input, sg_order, expected_result",
        [
            pytest.param(
                [
                    AhbLine(
                        segment_group_key="Foo",
                        guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Bar",
                        guid=uuid.UUID("9ec5ddb3-3721-48be-9d57-8742e08aa7cf"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key=None,
                        guid=uuid.UUID("7beb2471-0fd2-4b6b-8aae-a5d1f153972d"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Bar",
                        guid=uuid.UUID("365e1f9d-0bb9-43ab-921b-9ffa16b6df86"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Foo",
                        guid=uuid.UUID("8874a7c9-0143-4aa5-b7fe-5225bb25d2c5"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                ],
                ["Foo", "Bar", None],
                [
                    AhbLine(
                        segment_group_key="Foo",
                        guid=uuid.UUID("12b1a98a-edf5-4177-89e5-a6d8a92c5fdc"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Foo",
                        guid=uuid.UUID("8874a7c9-0143-4aa5-b7fe-5225bb25d2c5"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Bar",
                        guid=uuid.UUID("9ec5ddb3-3721-48be-9d57-8742e08aa7cf"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key="Bar",
                        guid=uuid.UUID("365e1f9d-0bb9-43ab-921b-9ffa16b6df86"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                    AhbLine(
                        segment_group_key=None,
                        guid=uuid.UUID("7beb2471-0fd2-4b6b-8aae-a5d1f153972d"),
                        segment_code=None,
                        data_element=None,
                        value_pool_entry=None,
                        ahb_expression=None,
                        name=None,
                    ),
                ],
            )
        ],
    )
    def test_sorted_segment_groups(
        self, unsorted_input: List[AhbLine], sg_order: List[Optional[str]], expected_result: List[AhbLine]
    ):
        actual = FlatAnwendungshandbuch._sorted_lines_by_segment_groups(unsorted_input, sg_order)
        assert actual == expected_result



from typing import Optional

import pytest  # type:ignore[import]

from maus.division_helper import Division, get_division, is_electricity_name, is_gas_name


class TestDivisionHelper:
    """
    Tests the behaviour of the Division Helper.
    """

    @pytest.mark.parametrize(
        "field_name, expected_is_strom",
        [
            pytest.param("Foo", None),
            pytest.param("Gas", False),
            pytest.param("Strom", True),
            pytest.param("Turnusableseintervall des MSB (Strom)", True),
            pytest.param("Geplante Turnusablesung des NB (Gas)", False),
        ],
    )
    def test_is_strom_field(self, field_name: str, expected_is_strom: Optional[bool]):
        actual = is_electricity_name(field_name)
        assert actual == expected_is_strom

    @pytest.mark.parametrize(
        "field_name, expected_is_gas",
        [
            pytest.param("Foo", None),
            pytest.param("Gas", True),
            pytest.param("Strom", False),
            pytest.param("Turnusableseintervall des MSB (Strom)", False),
            pytest.param("Geplante Turnusablesung des NB (Gas)", True),
        ],
    )
    def test_is_gas_field(self, field_name: str, expected_is_gas: Optional[bool]):
        actual = is_gas_name(field_name)
        assert actual == expected_is_gas

    @pytest.mark.parametrize(
        "field_name, expected_division",
        [
            pytest.param(None, None),
            pytest.param("", None),
            pytest.param("Foo", None),
            pytest.param("Gas", Division.GAS),
            pytest.param("Strom", Division.ELECTRICITY),
            pytest.param("Turnusableseintervall des MSB (Strom)", Division.ELECTRICITY),
            pytest.param("Geplante Turnusablesung des NB (Gas)", Division.GAS),
        ],
    )
    def test_get_division(self, field_name: Optional[str], expected_division: Optional[Division]):
        actual = get_division(field_name)
        assert actual == expected_division



import pytest  # type:ignore[import]

from maus.models.edifact_components import (
    DataElementDataType,
    DataElementFreeText,
    DataElementFreeTextSchema,
    DataElementValuePool,
    DataElementValuePoolSchema,
    EdifactStack,
    EdifactStackLevel,
    Segment,
    SegmentGroup,
    SegmentGroupSchema,
    SegmentSchema,
    ValuePoolEntry,
)
from unittests.serialization_test_helper import assert_serialization_roundtrip  # type:ignore[import]


class TestEdifactComponents:
    """
    Tests the behaviour of the Edifact Components
    """

    @pytest.mark.parametrize(
        "free_text, expected_json_dict",
        [
            pytest.param(
                DataElementFreeText(
                    ahb_expression="Muss [1]",
                    entered_input="Hello Maus",
                    discriminator="foo",
                    data_element_id="2222",
                    value_type=DataElementDataType.TEXT,
                ),
                {
                    "ahb_expression": "Muss [1]",
                    "entered_input": "Hello Maus",
                    "discriminator": "foo",
                    "data_element_id": "2222",
                    "value_type": "TEXT",
                },
            ),
        ],
    )
    def test_free_text_serialization_roundtrip(self, free_text: DataElementFreeText, expected_json_dict: dict):
        assert_serialization_roundtrip(free_text, DataElementFreeTextSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "value_pool, expected_json_dict",
        [
            pytest.param(
                DataElementValuePool(
                    value_pool=[
                        ValuePoolEntry(qualifier="hello", meaning="world", ahb_expression="X"),
                        ValuePoolEntry(qualifier="maus", meaning="rocks", ahb_expression="X"),
                    ],
                    discriminator="foo",
                    data_element_id="0022",
                ),
                {
                    "value_pool": [
                        {"qualifier": "hello", "meaning": "world", "ahb_expression": "X"},
                        {"qualifier": "maus", "meaning": "rocks", "ahb_expression": "X"},
                    ],
                    "discriminator": "foo",
                    "data_element_id": "0022",
                    "value_type": "VALUE_POOL",
                },
            ),
        ],
    )
    def test_value_pool_serialization_roundtrip(self, value_pool: DataElementValuePool, expected_json_dict: dict):
        assert_serialization_roundtrip(value_pool, DataElementValuePoolSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "segment, expected_json_dict",
        [
            pytest.param(
                Segment(
                    ahb_expression="X",
                    section_name="foo",
                    data_elements=[
                        DataElementValuePool(
                            value_pool=[
                                ValuePoolEntry(qualifier="hello", meaning="world", ahb_expression="X"),
                                ValuePoolEntry(qualifier="maus", meaning="rocks", ahb_expression="X"),
                            ],
                            discriminator="baz",
                            data_element_id="0329",
                        ),
                        DataElementFreeText(
                            ahb_expression="Muss [1]",
                            entered_input="Hello Maus",
                            discriminator="bar",
                            data_element_id="0330",
                        ),
                    ],
                    discriminator="foo",
                ),
                {
                    "ahb_expression": "X",
                    "section_name": "foo",
                    "data_elements": [
                        {
                            "value_pool": [
                                {"qualifier": "hello", "meaning": "world", "ahb_expression": "X"},
                                {"qualifier": "maus", "meaning": "rocks", "ahb_expression": "X"},
                            ],
                            "discriminator": "baz",
                            "data_element_id": "0329",
                            "value_type": "VALUE_POOL",
                        },
                        {
                            "ahb_expression": "Muss [1]",
                            "entered_input": "Hello Maus",
                            "discriminator": "bar",
                            "data_element_id": "0330",
                            "value_type": "TEXT",
                        },
                    ],
                    "discriminator": "foo",
                },
            ),
        ],
    )
    def test_segment_serialization_roundtrip(self, segment: Segment, expected_json_dict: dict):
        assert_serialization_roundtrip(segment, SegmentSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "segment_group, expected_json_dict",
        [
            pytest.param(
                SegmentGroup(
                    ahb_expression="expr A",
                    discriminator="disc A",
                    segments=[
                        Segment(
                            ahb_expression="expr B",
                            discriminator="disc B",
                            section_name="bar",
                            data_elements=[
                                DataElementValuePool(
                                    value_pool=[
                                        ValuePoolEntry(qualifier="hello", meaning="world", ahb_expression="X"),
                                        ValuePoolEntry(qualifier="maus", meaning="rocks", ahb_expression="X"),
                                    ],
                                    discriminator="baz",
                                    data_element_id="3333",
                                ),
                                DataElementFreeText(
                                    ahb_expression="Muss [1]",
                                    entered_input="Hello Maus",
                                    discriminator="bar",
                                    data_element_id="4444",
                                ),
                            ],
                        ),
                    ],
                    segment_groups=[
                        SegmentGroup(
                            discriminator="disc C",
                            ahb_expression="expr C",
                            segments=[
                                Segment(
                                    section_name="foo",
                                    ahb_expression="expr Y",
                                    discriminator="disc Y",
                                    data_elements=[],
                                )
                            ],
                            segment_groups=None,
                        ),
                    ],
                ),
                {
                    "ahb_expression": "expr A",
                    "discriminator": "disc A",
                    "segments": [
                        {
                            "section_name": "bar",
                            "ahb_expression": "expr B",
                            "discriminator": "disc B",
                            "data_elements": [
                                {
                                    "value_pool": [
                                        {"qualifier": "hello", "meaning": "world", "ahb_expression": "X"},
                                        {"qualifier": "maus", "meaning": "rocks", "ahb_expression": "X"},
                                    ],
                                    "discriminator": "baz",
                                    "data_element_id": "3333",
                                    "value_type": "VALUE_POOL",
                                },
                                {
                                    "ahb_expression": "Muss [1]",
                                    "entered_input": "Hello Maus",
                                    "discriminator": "bar",
                                    "data_element_id": "4444",
                                    "value_type": "TEXT",
                                },
                            ],
                        }
                    ],
                    "segment_groups": [
                        {
                            "ahb_expression": "expr C",
                            "discriminator": "disc C",
                            "segments": [
                                {
                                    "section_name": "foo",
                                    "ahb_expression": "expr Y",
                                    "discriminator": "disc Y",
                                    "data_elements": [],
                                }
                            ],
                            "segment_groups": None,
                        }
                    ],
                },
            ),
        ],
    )
    def test_segment_group_serialization_roundtrip(self, segment_group: SegmentGroup, expected_json_dict: dict):
        assert_serialization_roundtrip(segment_group, SegmentGroupSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "stack_x, stack_y, x_is_sub_stack_of_y, x_is_parent_of_y",
        [
            pytest.param(EdifactStack(levels=[]), EdifactStack(levels=[]), True, True, id="equality"),
            pytest.param(
                EdifactStack(levels=[EdifactStackLevel(name="x", is_groupable=False)]),
                EdifactStack(levels=[]),
                True,
                False,
                id="any stack is sub stack of root",
            ),
            pytest.param(
                EdifactStack(levels=[]),
                EdifactStack(levels=[EdifactStackLevel(name="x", is_groupable=False)]),
                False,
                True,
                id="other too deep",
            ),
            pytest.param(
                EdifactStack(levels=[EdifactStackLevel(name="x", is_groupable=True)]),
                EdifactStack(levels=[EdifactStackLevel(name="x", is_groupable=False)]),
                False,
                False,
                id="different groubability",
            ),
            pytest.param(
                EdifactStack(levels=[EdifactStackLevel(name="x", is_groupable=True)]),
                EdifactStack(levels=[EdifactStackLevel(name="y", is_groupable=True)]),
                False,
                False,
                id="different name",
            ),
            pytest.param(
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="a", is_groupable=True),
                        EdifactStackLevel(name="b", is_groupable=True),
                    ]
                ),
                EdifactStack(levels=[EdifactStackLevel(name="a", is_groupable=True)]),
                True,
                False,
                id="yes",
            ),
        ],
    )
    def test_edifact_stack_is_sub_or_parent_of(
        self, stack_x: EdifactStack, stack_y: EdifactStack, x_is_sub_stack_of_y: bool, x_is_parent_of_y: bool
    ):
        assert stack_x.is_sub_stack_of(stack_y) == x_is_sub_stack_of_y
        assert stack_x.is_parent_of(stack_y) == x_is_parent_of_y

    @pytest.mark.parametrize(
        "json_path",
        [
            pytest.param('$["foo"][0]["asd"]["bar"][29]["sss(asd)"][2]'),
            pytest.param('$["asd"][223239102]["Hallo ()asd90e34A)SD=A)D"]["bar"][29]["sss(asd)"][2]'),
        ],
    )
    def test_edifact_stack_to_from_json_path(self, json_path: str):
        stack = EdifactStack.from_json_path(json_path)
        assert stack.to_json_path() == json_path

    def test_segment_group_can_be_instantiated_without_explicitly_defining_sub_groups(self):
        """
        Tests https://github.com/Hochfrequenz/mig_ahb_utility_stack/issues/41
        """
        sg = SegmentGroup(
            discriminator="Foo",
            ahb_expression="",
        )  # must not throw an exception
        assert sg is not None
        assert sg.segment_groups is None
        assert sg.segments is None



from pathlib import Path

import pytest  # type:ignore[import]
from test_mig_xml_reader import ALL_MIG_XML_FILES

from maus import to_deep_ahb
from maus.deep_ahb_mig_joiner import replace_discriminators_with_edifact_stack
from maus.models.anwendungshandbuch import DeepAnwendungshandbuchSchema
from maus.models.message_implementation_guide import SegmentGroupHierarchySchema
from maus.reader.flat_ahb_reader import FlatAhbCsvReader
from maus.reader.mig_reader import MigXmlReader
from unittests.test_mig import ALL_SGH_FILES  # type:ignore[import]


class TestIntegration11042:
    """
    tests the whole MAUS package on the example of an 11042 AHB
    more of an integration test than a single unit
    """

    @ALL_SGH_FILES
    @ALL_MIG_XML_FILES
    @pytest.mark.datafiles("./unittests/ahbs/FV2204/UTILMD/11042.csv")
    @pytest.mark.datafiles("./unittests/ahbs/FV2204/UTILMD/11042_deep.json")
    def test_csv_file_reading_11042(self, datafiles):
        path_to_csv: Path = datafiles / "11042.csv"
        reader = FlatAhbCsvReader(file_path=path_to_csv)
        flat_ahb = reader.to_flat_ahb()
        with open(datafiles / "sgh_utilmd.json", "r", encoding="utf-8") as sgh_file:
            sgh = SegmentGroupHierarchySchema().loads(sgh_file.read())
        actual_deep_ahb = to_deep_ahb(flat_ahb, sgh)
        with open(datafiles / "11042_deep.json", "r", encoding="utf-8") as deep_ahb_file:
            expected_deep_ahb = DeepAnwendungshandbuchSchema().loads(deep_ahb_file.read())
        actual_json = DeepAnwendungshandbuchSchema().dumps(actual_deep_ahb, ensure_ascii=True, sort_keys=True)
        assert actual_deep_ahb == expected_deep_ahb
        return
        ## this happens later...
        mig_reader = MigXmlReader(Path(datafiles) / Path("utilmd.xml"))
        assert mig_reader is not None
        replace_discriminators_with_edifact_stack(actual_deep_ahb, mig_reader, ignore_errors=True)
        actual_maus_json = DeepAnwendungshandbuchSchema().dumps(actual_deep_ahb, ensure_ascii=True, sort_keys=True)
        with open(datafiles / "35001_maus.json", "r", encoding="utf-8") as maus_file:
            expected_maus = DeepAnwendungshandbuchSchema().loads(maus_file.read())
        assert actual_deep_ahb == expected_maus



import json
from typing import TypeVar

from marshmallow import Schema  # type:ignore[import]

T = TypeVar("T")


def assert_serialization_roundtrip(serializable_object: T, schema: Schema, expected_json_dict: dict) -> T:
    """
    Serializes the serializable_object using the provided schema,
    asserts, that the result is equal to the expected_json_dict
    then deserializes it again and asserts on equality with the original serializable_object

    :returns the deserialized_object
    """
    json_string = schema.dumps(serializable_object)
    assert json_string is not None
    actual_json_dict = json.loads(json_string)
    assert actual_json_dict == expected_json_dict
    deserialized_object = schema.loads(json_data=json_string)
    assert isinstance(deserialized_object, type(serializable_object))
    assert deserialized_object == serializable_object
    return deserialized_object
